from collections import Counter
from typing import Any, List

from mlflow.store.artifact.models_artifact_repo import ModelsArtifactRepository
from mlflow.store.artifact.runs_artifact_repo import RunsArtifactRepository


def is_artifact_uri(uri):
    """
    Checks the artifact URI is associated with a MLflow model or run.
    The actual URI can be a model URI, model URI + subdirectory, or model URI + path to artifact file.
    """
    return ModelsArtifactRepository.is_models_uri(
        uri
    ) or RunsArtifactRepository.is_runs_uri(uri)


def get_duplicates(elements: List[Any]) -> List[Any]:
    """
    Returns duplicate elements in the order they first appear.
    """
    element_counts = Counter(elements)
    duplicates = []
    for e in element_counts.keys():
        if element_counts[e] > 1:
            duplicates.append(e)
    return duplicates


def validate_strings_unique(strings: List[str], error_template: str):
    """
    Validates all strings are unique, otherwise raise ValueError with the error template and duplicates.
    Passes single-quoted, comma delimited duplicates to the error template.
    """
    duplicate_strings = get_duplicates(strings)
    if duplicate_strings:
        duplicates_formatted = ", ".join([f"'{s}'" for s in duplicate_strings])
        raise ValueError(error_template.format(duplicates_formatted))


def get_unique_list_order(elements: List[Any]) -> List[Any]:
    """
    Returns unique elements in the order they first appear.
    """
    return list(dict.fromkeys(elements))



# Shared mlflow model constants that are used in both the Feature Store client and lookup client

# Module name of the original mlflow_model
MLFLOW_MODEL_NAME = "databricks.feature_store.mlflow_model"

# FeatureStoreClient.log_model will log models containing a 'raw_model' folder within the data_path.
# This folder stores the MLmodel for the raw model, which is needed to run inference.
# FeatureStoreClient.log_model may be called by any version of the FeatureStoreClient >= 0.3.0,
# however this module may be from a more recent version. As such:
# *** The value of this constant SHOULD NEVER CHANGE ***
RAW_MODEL_FOLDER = "raw_model"

# Constant for the ML model filename
ML_MODEL = "MLmodel"

# The package name of the feature lookup client as published on PyPI
FEATURE_LOOKUP_CLIENT_PIP_PACKAGE = "databricks-feature-lookup"

# The major version of the feature lookup client.  This is a shared constant in order to tightly couple:
#
# - The Feature lookup client major version (eg `{FEATURE_LOOKUP_CLIENT_PIP_PACKAGE}.0.1`)
# - The pinned major version of the databricks-feature-lookup requirement in logged model's conda.yaml
#
# If needed, these can be decoupled into separate constants, but that decision should be carefully considered.
FEATURE_LOOKUP_CLIENT_MAJOR_VERSION = 0

# The minor.micro version number of the feature lookup client Alpha version. The Alpha version is used
# by core client when logging model to create Feature Serving Endpoint.
# Minor version
FEATURE_LOOKUP_CLIENT_ALPHA_MINOR_VERSION = 3
# Micro version
FEATURE_LOOKUP_CLIENT_ALPHA_MICRO_VERSION = 1

DATABRICKS = "Databricks"
SAGEMAKER = "SageMaker"



import os
from typing import Dict, List, Optional, Tuple, Union

import pandas as pd

from databricks.feature_store.entities.online_feature_table import (
    AbstractOnlineFeatureTable,
    OnlineFeatureTable,
    OnlineFeatureTableForSageMakerServing,
)
from databricks.feature_store.entities.query_mode import QueryMode
from databricks.feature_store.entities.store_type import StoreType
from databricks.feature_store.lookup_engine import (
    AwsAccessKey,
    LookupCosmosDbEngine,
    LookupDynamoDbEngine,
    LookupEngine,
    LookupMySqlEngine,
    LookupSqlEngine,
    LookupSqlServerEngine,
)
from databricks.feature_store.mlflow_model_constants import DATABRICKS, SAGEMAKER
from databricks.feature_store.utils.metrics_utils import LookupClientMetrics

# The provisioner of this model is expected to set the following environment variable for each
# feature table if the feature store is SQL based:
#  (1) <online_store_for_serving.read_secret_prefix>_USER
#  (2) <online_store_for_serving.read_secret_prefix>_PASSWORD
USER_SUFFIX = "_USER"
PASSWORD_SUFFIX = "_PASSWORD"
# For DynamoDB the following variables should be set:
#  (1) <online_store_for_serving.read_secret_prefix>_ACCESS_KEY_ID
#  (2) <online_store_for_serving.read_secret_prefix>_SECRET_ACCESS_KEY
ACCESS_KEY_ID_SUFFIX = "_ACCESS_KEY_ID"
SECRET_ACCESS_KEY_SUFFIX = "_SECRET_ACCESS_KEY"
# For Cosmos DB the following variable should be set:
#  (1) <online_store_for_serving.read_secret_prefix>_AUTHORIZATION_KEY
AUTHORIZATION_KEY_SUFFIX = "_AUTHORIZATION_KEY"
LookupKeyType = Tuple[str, ...]


def generate_lookup_sql_engine(
    online_feature_table: Union[
        AbstractOnlineFeatureTable, List[AbstractOnlineFeatureTable]
    ],
    creds: Tuple[str, str],
) -> LookupSqlEngine:
    if isinstance(online_feature_table, List):
        raise Exception(f"Internal error: Batch lookup is unsupported in SQL Engine.")

    ro_user, ro_password = creds
    if online_feature_table.online_store.store_type == StoreType.SQL_SERVER:
        return LookupSqlServerEngine(online_feature_table, ro_user, ro_password)
    return LookupMySqlEngine(online_feature_table, ro_user, ro_password)


def generate_lookup_dynamodb_engine(
    online_feature_table: Union[
        AbstractOnlineFeatureTable, List[AbstractOnlineFeatureTable]
    ],
    creds: Optional[Tuple[str, str]],
) -> LookupDynamoDbEngine:
    if creds:
        access_key_id, secret_access_key = creds

        return LookupDynamoDbEngine(
            online_feature_table,
            access_key=AwsAccessKey(access_key_id, secret_access_key),
        )
    else:
        return LookupDynamoDbEngine(online_feature_table, access_key=None)


def generate_lookup_cosmosdb_engine(
    online_feature_table: Union[
        AbstractOnlineFeatureTable, List[AbstractOnlineFeatureTable]
    ],
    creds: str,
) -> LookupCosmosDbEngine:
    if isinstance(online_feature_table, List):
        raise Exception(f"Internal error: Batch lookup is unsupported in Cosmos DB.")
    return LookupCosmosDbEngine(online_feature_table, authorization_key=creds)


def load_credentials_from_env(online_ft: OnlineFeatureTable):
    def get_env_var(env_var_suffix: str) -> str:
        env_var = online_ft.online_store.read_secret_prefix + env_var_suffix
        if env_var not in os.environ:
            raise Exception(
                f"Internal error: {env_var} not found for feature table {online_ft.feature_table_name}."
            )
        return os.getenv(env_var)

    if online_ft.online_store.store_type == StoreType.DYNAMODB:
        return (
            get_env_var(ACCESS_KEY_ID_SUFFIX),
            get_env_var(SECRET_ACCESS_KEY_SUFFIX),
        )
    elif online_ft.online_store.store_type == StoreType.COSMOSDB:
        return get_env_var(AUTHORIZATION_KEY_SUFFIX)
    else:
        return (
            get_env_var(USER_SUFFIX),
            get_env_var(PASSWORD_SUFFIX),
        )


# TODO(ML-26146): refactor this method and above helpers to appropriate seperate util files
def tables_share_dynamodb_access_keys(
    online_feature_tables: List[AbstractOnlineFeatureTable],
):
    # check region and credentials to see if all feature tables are using same DynamoDB accesss keys
    region, creds = None, None
    for ft in online_feature_tables:
        if ft.online_store.store_type != StoreType.DYNAMODB:
            return False
        elif region == None:
            region = ft.online_store.extra_configs.region
            creds = load_credentials_from_env(ft)
        elif (
            region != ft.online_store.extra_configs.region
            or load_credentials_from_env(ft) != creds
        ):
            return False
    return True


def is_primary_key_lookup(
    online_feature_tables: List[AbstractOnlineFeatureTable],
):
    for oft in online_feature_tables:
        mode = oft.online_store.query_mode
        if mode == QueryMode.RANGE_QUERY:
            return False
        elif mode == QueryMode.PRIMARY_KEY_LOOKUP:
            continue
        else:
            raise ValueError(f"Unsupported query mode: {mode}")
    return True


def check_instance_type(
    online_feature_table: Union[
        AbstractOnlineFeatureTable, List[AbstractOnlineFeatureTable]
    ],
    cls: Union[OnlineFeatureTable, OnlineFeatureTableForSageMakerServing],
    serving_environment: str,
):
    if isinstance(online_feature_table, cls):
        return True
    elif isinstance(online_feature_table, List):
        for oft in online_feature_table:
            if not isinstance(oft, cls):
                raise Exception(
                    f"Internal Error: Expected {cls} when serving_environment={serving_environment}"
                    f". Got {type(oft)}."
                )
        return True

    raise Exception(
        f"Internal Error: Expected {cls} when serving_environment={serving_environment}"
        f". Got {type(online_feature_table)}."
    )


class OnlineLookupClient:
    def __init__(
        self,
        online_feature_table: Union[
            AbstractOnlineFeatureTable, List[AbstractOnlineFeatureTable]
        ],
        serving_environment: str,
    ):
        self.lookup_engine = self._generate_lookup_engine(
            online_feature_table, serving_environment
        )

    @staticmethod
    def _generate_lookup_engine(
        online_feature_table: Union[
            AbstractOnlineFeatureTable, List[AbstractOnlineFeatureTable]
        ],
        serving_environment: str,
    ):
        if serving_environment == DATABRICKS:
            check_instance_type(
                online_feature_table, OnlineFeatureTable, serving_environment
            )
            return OnlineLookupClient._generate_lookup_engine_databricks(
                online_feature_table
            )
        elif serving_environment == SAGEMAKER:
            check_instance_type(
                online_feature_table,
                OnlineFeatureTableForSageMakerServing,
                serving_environment,
            )
            return OnlineLookupClient._generate_lookup_engine_sagemaker(
                online_feature_table
            )

        raise Exception(
            f"Internal Error: Unexpected serving_environment {serving_environment}."
        )

    @classmethod
    def _generate_lookup_engine_databricks(
        cls,
        online_feature_table: Union[
            AbstractOnlineFeatureTable, List[AbstractOnlineFeatureTable]
        ],
    ):
        first_online_feature_table = None
        if isinstance(online_feature_table, List):
            if len(online_feature_table) == 0:
                raise Exception(
                    f"Internal Error: No feature table passed while creating lookup engine."
                )
            first_online_feature_table = online_feature_table[0]
        else:
            first_online_feature_table = online_feature_table

        creds = load_credentials_from_env(first_online_feature_table)
        if first_online_feature_table.online_store.store_type == StoreType.DYNAMODB:
            return generate_lookup_dynamodb_engine(online_feature_table, creds)
        elif first_online_feature_table.online_store.store_type == StoreType.COSMOSDB:
            return generate_lookup_cosmosdb_engine(online_feature_table, creds)
        else:
            return generate_lookup_sql_engine(online_feature_table, creds)

    @classmethod
    def _generate_lookup_engine_sagemaker(
        cls, online_feature_table: OnlineFeatureTableForSageMakerServing
    ):
        return generate_lookup_dynamodb_engine(online_feature_table, creds=None)

    def batch_lookup_features(
        self,
        lookup_df_dict: Dict[str, Dict[LookupKeyType, pd.DataFrame]],
        feature_names_dict: Dict[str, Dict[LookupKeyType, List[str]]],
        *,
        metrics: LookupClientMetrics = None,
    ) -> Dict[str, Dict[LookupKeyType, pd.DataFrame]]:
        """
        :param lookup_df_dict: dictionary from online feature table name to primary key dfs.
        :param feature_names_dict:  dictionary from online feature table name to feature names.
        :return a Dictionary from online feature table name to feature values.
        """
        return self.lookup_engine.batch_lookup_features(
            lookup_df_dict, feature_names_dict, metrics=metrics
        )

    def lookup_features(
        self,
        lookup_df: pd.DataFrame,
        feature_names: List[str],
        *,
        metrics: LookupClientMetrics = None,
    ) -> pd.DataFrame:
        """Uses a Python database connection to lookup features in feature_names using
        the lookup keys and values in lookup_df. The online store database and table name are
        obtained from the AbstractOnlineFeatureTable passed to the constructor.

        The resulting DataFrame has the same number of rows as lookup_df. In the case that a lookup
        key cannot be found, a row of NaNs will be returned in the resulting DataFrame.

        Throws an exception if the table, lookup keys, or feature columns do not exist in the
        online store.

        :param lookup_df: Pandas DataFrame containing lookup keys and values. The DataFrame should
        contain one column for each primary key of the online feature table, and one row for each
        entity to look up.
        :param feature_names: A list of feature names to look up.
        :return: Pandas DataFrame containing feature keys and values fetched from the online store.
        """
        features = self.lookup_engine.lookup_features(
            lookup_df, feature_names, metrics=metrics
        )
        return features

    def cleanup(self):
        """
        Performs any cleanup associated with the online store.
        :return:
        """
        self.lookup_engine.shutdown()



"""
Defines the LookupCosmosDbEngine class, which is used to perform lookups on Cosmos DB online stores.
"""

import functools
import logging
from typing import Any, List

import numpy as np
import pandas as pd
from azure.cosmos import CosmosClient
from azure.cosmos.exceptions import CosmosResourceNotFoundError

from databricks.feature_store.entities.online_feature_table import OnlineFeatureTable
from databricks.feature_store.entities.query_mode import QueryMode
from databricks.feature_store.lookup_engine.lookup_engine import LookupEngine
from databricks.feature_store.utils.cosmosdb_type_utils import (
    COSMOSDB_DATA_TYPE_CONVERTER_FACTORY,
)
from databricks.feature_store.utils.cosmosdb_utils import (
    FEATURE_STORE_SENTINEL_ID_VALUE,
    PARTITION_KEY,
    PATHS,
    PRIMARY_KEY_PROPERTY_NAME_VALUE,
    is_read_item_not_found_error,
    to_cosmosdb_primary_key,
)
from databricks.feature_store.utils.metrics_utils import (
    NAN_FEATURE_COUNT,
    LookupClientMetrics,
    lookup_call_maybe_with_metrics,
)

_logger = logging.getLogger(__name__)


class LookupCosmosDbEngine(LookupEngine):
    def __init__(
        self, online_feature_table: OnlineFeatureTable, authorization_key: str
    ):
        """
        :param online_feature_table: OnlineFeatureTable to look up feature values from.
        :param authorization_key: Uses this authorization key to authenticate with Cosmos DB.
        """
        # The online feature table name is of the format database_name.container_name.
        self.online_table_name = online_feature_table.online_feature_table_name
        self.database_name, self.container_name = self.online_table_name.split(".")

        # Retrieve the relevant configuration and helpers needed for lookup.
        self.account_uri = online_feature_table.online_store.extra_configs.account_uri
        self.query_mode = online_feature_table.online_store.query_mode
        self.timestamp_keys = online_feature_table.timestamp_keys
        self.primary_keys_to_type_converter = {
            pk.name: COSMOSDB_DATA_TYPE_CONVERTER_FACTORY.get_converter(pk)
            for pk in online_feature_table.primary_keys
        }
        self.features_to_type_converter = {
            feature.name: COSMOSDB_DATA_TYPE_CONVERTER_FACTORY.get_converter(feature)
            for feature in online_feature_table.features
        }

        # Initialize the CosmosClient used for lookup.
        self._client = CosmosClient(self.account_uri, authorization_key)
        self._database_client = self._client.get_database_client(self.database_name)
        self._container_client = self._database_client.get_container_client(
            self.container_name
        )
        self._validate_online_feature_table()

    def _validate_online_feature_table(
        self,
    ) -> None:
        # Check that the online container exists and retrieve its details.
        # Otherwise, a CosmosResourceNotFoundError exception is thrown.
        container_desc = self._container_client.read()

        # All container descriptions contain the partition key and paths. Check the partition key is as expected.
        partition_key_paths = container_desc[PARTITION_KEY][PATHS]
        if partition_key_paths != [f"/{PRIMARY_KEY_PROPERTY_NAME_VALUE}"]:
            raise ValueError(
                f"Online Table {self.online_table_name} primary key schema is not configured properly."
            )

    def lookup_features(
        self,
        lookup_df: pd.DataFrame,
        feature_names: List[str],
        *,
        metrics: LookupClientMetrics = None,
    ) -> pd.DataFrame:
        query = functools.partial(
            self._run_lookup_cosmosdb_query, feature_names, metrics=metrics
        )
        feature_df = lookup_df.apply(query, axis=1, result_type="expand")
        feature_df.columns = feature_names
        return feature_df

    def _lookup_primary_key(self, cosmosdb_primary_key: str):
        try:
            # The expected response is the item with additional system properties, e.g. {"feat1": ..., "sys1": ...}
            # It's not possible to selectively retrieve features when using point reads (`client.read_item`).
            # However, point reads are recommended over queries (`client.query_items`) for performance and cost.
            # https://docs.microsoft.com/en-us/azure/cosmos-db/optimize-cost-reads-writes

            # As of azure-cosmos==4.2.0 (the minimum required version), `client.read_item` retries up to either of:
            # 30 seconds of timeout or 9 total attempts. Thus, no manual retry handling needs to be done.
            return self._container_client.read_item(
                item=FEATURE_STORE_SENTINEL_ID_VALUE, partition_key=cosmosdb_primary_key
            )
        except Exception as e:
            # Return None if the error was caused by read_item finding no result. Otherwise, re-raise the error.
            if is_read_item_not_found_error(e):
                return None
            _logger.warning(
                f"Encountered error reading from {self.online_table_name}:\n{e}"
            )
            raise e

    def _run_lookup_cosmosdb_query(
        self,
        feature_names: List[str],
        lookup_row: pd.core.series.Series,
        *,
        metrics: LookupClientMetrics = None,
    ):
        """
        This helper function executes a single Cosmos DB query.
        """
        nan_features_count = 0
        cosmosdb_lookup_row = self._pandas_to_cosmosdb(lookup_row)
        cosmosdb_primary_key = to_cosmosdb_primary_key(cosmosdb_lookup_row)
        if self.query_mode == QueryMode.PRIMARY_KEY_LOOKUP:
            lookup_pk = lookup_call_maybe_with_metrics(
                self._lookup_primary_key, metrics
            )
            feature_values = lookup_pk(cosmosdb_primary_key)
        else:
            raise ValueError(f"Unsupported query mode: {self.query_mode}")

        if not feature_values:
            _logger.warning(
                f"No feature values found in {self.online_table_name} for {cosmosdb_lookup_row}."
            )
            nan_features_count = len(feature_names)
            return np.full(len(feature_names), np.nan)

        results = [feature_values.get(f, np.nan) for f in feature_names]
        nan_features_count = sum([pd.isnull(x) for x in results])
        if metrics:
            # 0 nan feature count still needs to be recorded to propogate metrics for inference request
            metrics.increase_metric(NAN_FEATURE_COUNT, nan_features_count)

        # Return the result
        return self._cosmosdb_to_pandas(results, feature_names)

    def _pandas_to_cosmosdb(self, row: pd.core.series.Series) -> List[Any]:
        """
        Converts the input Pandas row to the Cosmos DB online equivalent Python types.
        """
        return [
            self.primary_keys_to_type_converter[pk_name].to_online_store(pk_value)
            for pk_name, pk_value in row.items()
        ]

    def _cosmosdb_to_pandas(
        self, results: List[Any], feature_names: List[str]
    ) -> List[Any]:
        """
        Converts the online store retrieved item to Pandas compatible Python values for the given features.
        """
        feature_names_and_values = zip(feature_names, results)
        return [
            self.features_to_type_converter[feature_name].to_pandas(feature_value)
            for feature_name, feature_value in feature_names_and_values
        ]

    def shutdown(self) -> None:
        """
        Cleans up the store connection if it exists on the Cosmos DB online store.
        :return:
        """
        # Cosmos DB connections are stateless http connections and hence do not need an explicit
        # shutdown operation.
        pass



""" Defines the type conversion classes from pandas to dynamodb and vice versa.
"""
from databricks.feature_store.entities.data_type import DataType
from databricks.feature_store.utils.converter_utils import ConverterFactory
from databricks.feature_store.utils.pandas_type_utils import (
    PandasArrayTypeConverter,
    PandasBinaryTypeConverter,
    PandasDecimalTypeConverter,
    PandasDoubleTypeConverter,
    PandasEpochDateTypeConverter,
    PandasEpochTimestampTypeUsConverter,
    PandasFloatTypeConverter,
    PandasIntTypeConverter,
    PandasLongTypeConverter,
    PandasMapTypeConverter,
    PandasNumericBooleanConverter,
    PandasShortTypeConverter,
    PandasStringTypeConverter,
)

"""
All supported converters.
See the link below for relevant design decision on DynamoDB type conversion.
https://docs.google.com/document/d/1CvdYWUDqEsv69YVv1S9Co2vx-akEaki3v_H3Q2ZMDmo/edit#bookmark=id.qdunvvvuke0e
"""
BASIC_DATA_TYPE_CONVERTERS = {
    DataType.SHORT: PandasShortTypeConverter,
    DataType.INTEGER: PandasIntTypeConverter,
    DataType.LONG: PandasLongTypeConverter,
    DataType.FLOAT: PandasFloatTypeConverter,
    DataType.DOUBLE: PandasDoubleTypeConverter,
    DataType.BOOLEAN: PandasNumericBooleanConverter,
    DataType.STRING: PandasStringTypeConverter,
    DataType.TIMESTAMP: PandasEpochTimestampTypeUsConverter,
    DataType.DATE: PandasEpochDateTypeConverter,
    DataType.BINARY: PandasBinaryTypeConverter,
}

COMPLEX_DATA_TYPE_CONVERTERS = {
    DataType.DECIMAL: PandasDecimalTypeConverter,
    DataType.ARRAY: PandasArrayTypeConverter,
    DataType.MAP: PandasMapTypeConverter,
}

DYNAMODB_DATA_TYPE_CONVERTER_FACTORY = ConverterFactory(
    BASIC_DATA_TYPE_CONVERTERS, COMPLEX_DATA_TYPE_CONVERTERS
)



import pytest

from databricks.feature_store.entities.data_type import DataType
from databricks.feature_store.entities.online_feature_table import (
    FeatureDetails,
    OnlineFeatureTableForSageMakerServing,
    PrimaryKeyDetails,
    TimestampKeyDetails,
)
from databricks.feature_store.entities.online_store_for_serving import (
    DynamoDbConf,
    OnlineStoreForSageMakerServing,
)
from databricks.feature_store.entities.query_mode import QueryMode
from databricks.feature_store.protos.feature_store_serving_pb2 import (
    DataType as ProtoDataType,
)
from databricks.feature_store.protos.feature_store_serving_pb2 import (
    DynamoDbConf as ProtoDynamoDbConf,
)
from databricks.feature_store.protos.feature_store_serving_pb2 import (
    FeatureDetails as ProtoFeatureDetails,
)
from databricks.feature_store.protos.feature_store_serving_pb2 import (
    FeatureTablesForSageMakerServing as ProtoFeatureTablesForSageMakerServing,
)
from databricks.feature_store.protos.feature_store_serving_pb2 import (
    OnlineFeatureTableForSageMakerServing as ProtoOnlineFeatureTableForSageMakerServing,
)
from databricks.feature_store.protos.feature_store_serving_pb2 import (
    OnlineStoreForSageMakerServing as ProtoOnlineStoreForSageMakerServing,
)
from databricks.feature_store.protos.feature_store_serving_pb2 import (
    PrimaryKeyDetails as ProtoPrimaryKeyDetails,
)
from databricks.feature_store.protos.feature_store_serving_pb2 import (
    QueryMode as ProtoQueryMode,
)
from databricks.feature_store.protos.feature_store_serving_pb2 import (
    TimestampKeyDetails as ProtoTimestampKeyDetails,
)


@pytest.fixture(scope="function")
def sagemaker_online_store_fixture():
    return OnlineStoreForSageMakerServing(
        creation_timestamp_ms=123,
        extra_configs=DynamoDbConf(region="us-south-16"),
        query_mode=QueryMode.PRIMARY_KEY_LOOKUP,
    )


def validate_sagemaker_online_store_proto(the_proto):
    assert isinstance(the_proto, ProtoOnlineStoreForSageMakerServing)
    assert the_proto.creation_timestamp_ms == 123
    assert isinstance(the_proto.dynamodb_conf, ProtoDynamoDbConf)
    assert the_proto.dynamodb_conf.region == "us-south-16"
    assert the_proto.query_mode == QueryMode.PRIMARY_KEY_LOOKUP


@pytest.fixture(scope="function")
def sagemaker_online_store_proto_fixture():
    return ProtoOnlineStoreForSageMakerServing(
        creation_timestamp_ms=123,
        dynamodb_conf=ProtoDynamoDbConf(region="us-south-16"),
        query_mode=ProtoQueryMode.PRIMARY_KEY_LOOKUP,
    )


def validate_sagemaker_online_store(online_store):
    assert online_store.creation_timestamp_ms == 123
    assert isinstance(online_store.extra_configs, DynamoDbConf)
    assert online_store.extra_configs.region == "us-south-16"
    assert online_store.query_mode == QueryMode.PRIMARY_KEY_LOOKUP


@pytest.fixture(scope="function")
def sagemaker_online_feature_table_fixture(sagemaker_online_store_fixture):
    return OnlineFeatureTableForSageMakerServing(
        feature_table_name="test_db.test_ft",
        online_feature_table_name="test_db_online.test_ft_online",
        online_store=sagemaker_online_store_fixture,
        primary_keys=[
            PrimaryKeyDetails(name="pk1", data_type=DataType.INTEGER),
            PrimaryKeyDetails(name="pk2", data_type=DataType.STRING),
        ],
        feature_table_id="abc123",
        features=[
            FeatureDetails(name="Ft1", data_type=DataType.STRING),
            FeatureDetails(
                name="Ft2",
                data_type=DataType.ARRAY,
                data_type_details="""{"containsNull":true,"elementType":"integer","type":"array"}""",
            ),
        ],
        timestamp_keys=[TimestampKeyDetails(name="ts", data_type=DataType.TIMESTAMP)],
    )


def validate_sagemaker_online_feature_table_proto(the_proto):
    assert isinstance(the_proto, ProtoOnlineFeatureTableForSageMakerServing)
    assert the_proto.feature_table_name == "test_db.test_ft"
    assert the_proto.online_feature_table_name == "test_db_online.test_ft_online"
    assert len(the_proto.primary_keys) == 2
    primary_key_1 = the_proto.primary_keys[0]
    assert isinstance(primary_key_1, ProtoPrimaryKeyDetails)
    assert primary_key_1.name == "pk1"
    assert primary_key_1.data_type == DataType.INTEGER
    primary_key_2 = the_proto.primary_keys[1]
    assert isinstance(primary_key_2, ProtoPrimaryKeyDetails)
    assert primary_key_2.name == "pk2"
    assert primary_key_2.data_type == DataType.STRING
    assert len(the_proto.features) == 2
    ft_1 = the_proto.features[0]
    assert isinstance(ft_1, ProtoFeatureDetails)
    assert ft_1.name == "Ft1"
    assert ft_1.data_type == DataType.STRING
    ft_2 = the_proto.features[1]
    assert isinstance(ft_2, ProtoFeatureDetails)
    assert ft_2.name == "Ft2"
    assert ft_2.data_type == DataType.ARRAY
    assert (
        ft_2.data_type_details
        == """{"containsNull":true,"elementType":"integer","type":"array"}"""
    )
    assert len(the_proto.timestamp_keys) == 1
    assert the_proto.timestamp_keys[0] == ProtoTimestampKeyDetails(
        name="ts", data_type=ProtoDataType.TIMESTAMP
    )
    validate_sagemaker_online_store_proto(the_proto.online_store)


@pytest.fixture(scope="function")
def sagemaker_online_feature_table_proto_fixture(sagemaker_online_store_proto_fixture):
    the_proto = ProtoOnlineFeatureTableForSageMakerServing(
        feature_table_name="test_db.test_ft",
        online_feature_table_name="test_db_online.test_ft_online",
        primary_keys=[
            ProtoPrimaryKeyDetails(name="pk1", data_type=DataType.INTEGER),
            ProtoPrimaryKeyDetails(name="pk2", data_type=DataType.STRING),
        ],
        feature_table_id="abc123",
        features=[
            ProtoFeatureDetails(name="Ft1", data_type=DataType.STRING),
            ProtoFeatureDetails(
                name="Ft2",
                data_type=DataType.ARRAY,
                data_type_details="""{"containsNull":true,"elementType":"integer","type":"array"}""",
            ),
        ],
        timestamp_keys=[
            ProtoTimestampKeyDetails(name="ts", data_type=DataType.TIMESTAMP)
        ],
    )
    the_proto.online_store.CopyFrom(sagemaker_online_store_proto_fixture)
    return the_proto


def validate_sagemaker_online_feature_table(
    online_feature_table, feature_table_name="test_db.test_ft"
):
    assert online_feature_table.feature_table_name == feature_table_name
    assert (
        online_feature_table.online_feature_table_name
        == "test_db_online.test_ft_online"
    )
    assert online_feature_table._feature_table_id == "abc123"
    assert len(online_feature_table.primary_keys) == 2
    primary_key_1 = online_feature_table.primary_keys[0]
    assert primary_key_1.name == "pk1"
    assert primary_key_1.data_type == DataType.INTEGER
    primary_key_2 = online_feature_table.primary_keys[1]
    assert primary_key_2.name == "pk2"
    assert primary_key_2.data_type == DataType.STRING
    ft_1 = online_feature_table.features[0]
    assert ft_1.name == "Ft1"
    assert ft_1.data_type == DataType.STRING
    ft_2 = online_feature_table.features[1]
    assert ft_2.name == "Ft2"
    assert ft_2.data_type == DataType.ARRAY
    assert (
        ft_2.data_type_details
        == """{"containsNull":true,"elementType":"integer","type":"array"}"""
    )
    assert online_feature_table.timestamp_keys == [
        TimestampKeyDetails(name="ts", data_type=DataType.TIMESTAMP)
    ]
    assert isinstance(online_feature_table.online_store, OnlineStoreForSageMakerServing)
    assert isinstance(online_feature_table.online_store.extra_configs, DynamoDbConf)
    assert online_feature_table.online_store.extra_configs.region == "us-south-16"


@pytest.fixture(scope="function")
def feature_tables_for_sagemaker_serving_proto_fixture(
    sagemaker_online_feature_table_proto_fixture,
):
    the_proto = ProtoFeatureTablesForSageMakerServing()
    the_proto.online_tables.append(sagemaker_online_feature_table_proto_fixture)
    return the_proto



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: feature_store_serving.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from mlflow.protos.scalapb import scalapb_pb2 as scalapb_dot_scalapb__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1b\x66\x65\x61ture_store_serving.proto\x12\x12\x66\x65\x61turestorecommon\x1a\x15scalapb/scalapb.proto\"\xdc\x03\n\x15OnlineStoreForServing\x12(\n\x05\x63loud\x18\x01 \x01(\x0e\x32\x19.featurestorecommon.Cloud\x12\x31\n\nstore_type\x18\x02 \x01(\x0e\x32\x1d.featurestorecommon.StoreType\x12\x1a\n\x12read_secret_prefix\x18\x03 \x01(\t\x12\x1d\n\x15\x63reation_timestamp_ms\x18\x04 \x01(\x03\x12\x33\n\nmysql_conf\x18\x05 \x01(\x0b\x32\x1d.featurestorecommon.MySqlConfH\x00\x12<\n\x0fsql_server_conf\x18\x06 \x01(\x0b\x32!.featurestorecommon.SqlServerConfH\x00\x12\x39\n\rdynamodb_conf\x18\x07 \x01(\x0b\x32 .featurestorecommon.DynamoDbConfH\x00\x12\x39\n\rcosmosdb_conf\x18\x08 \x01(\x0b\x32 .featurestorecommon.CosmosDbConfH\x00\x12\x31\n\nquery_mode\x18\x15 \x01(\x0e\x32\x1d.featurestorecommon.QueryModeB\x0f\n\rextra_configs\"\'\n\tMySqlConf\x12\x0c\n\x04host\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\x05\"+\n\rSqlServerConf\x12\x0c\n\x04host\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\x05\"\x1e\n\x0c\x44ynamoDbConf\x12\x0e\n\x06region\x18\x01 \x01(\t\"#\n\x0c\x43osmosDbConf\x12\x13\n\x0b\x61\x63\x63ount_uri\x18\x01 \x01(\t\"j\n\x0e\x46\x65\x61tureDetails\x12\x0c\n\x04name\x18\x01 \x01(\t\x12/\n\tdata_type\x18\x02 \x01(\x0e\x32\x1c.featurestorecommon.DataType\x12\x19\n\x11\x64\x61ta_type_details\x18\x03 \x01(\t\"R\n\x11PrimaryKeyDetails\x12\x0c\n\x04name\x18\x01 \x01(\t\x12/\n\tdata_type\x18\x02 \x01(\x0e\x32\x1c.featurestorecommon.DataType\"T\n\x13TimestampKeyDetails\x12\x0c\n\x04name\x18\x01 \x01(\t\x12/\n\tdata_type\x18\x02 \x01(\x0e\x32\x1c.featurestorecommon.DataType\"\xe2\x02\n\x12OnlineFeatureTable\x12\x1a\n\x12\x66\x65\x61ture_table_name\x18\x01 \x01(\t\x12!\n\x19online_feature_table_name\x18\x02 \x01(\t\x12?\n\x0conline_store\x18\x03 \x01(\x0b\x32).featurestorecommon.OnlineStoreForServing\x12;\n\x0cprimary_keys\x18\x04 \x03(\x0b\x32%.featurestorecommon.PrimaryKeyDetails\x12\x18\n\x10\x66\x65\x61ture_table_id\x18\x05 \x01(\t\x12\x34\n\x08\x66\x65\x61tures\x18\x06 \x03(\x0b\x32\".featurestorecommon.FeatureDetails\x12?\n\x0etimestamp_keys\x18\x07 \x03(\x0b\x32\'.featurestorecommon.TimestampKeyDetails\"X\n\x17\x46\x65\x61tureTablesForServing\x12=\n\ronline_tables\x18\x01 \x03(\x0b\x32&.featurestorecommon.OnlineFeatureTable\"+\n\rMySqlMetadata\x12\x0c\n\x04host\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\x05\"/\n\x11SqlServerMetadata\x12\x0c\n\x04host\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\x05\"B\n\x10\x44ynamoDbMetadata\x12\x0e\n\x06region\x18\x01 \x01(\t\x12\x11\n\ttable_arn\x18\x02 \x01(\t\x12\x0b\n\x03ttl\x18\x03 \x01(\x03\"K\n\x10\x43osmosDbMetadata\x12\x13\n\x0b\x61\x63\x63ount_uri\x18\x01 \x01(\t\x12\x15\n\rcontainer_uri\x18\x02 \x01(\t\x12\x0b\n\x03ttl\x18\x03 \x01(\x03\"\xa9\x04\n\x0bOnlineStore\x12(\n\x05\x63loud\x18\x01 \x01(\x0e\x32\x19.featurestorecommon.Cloud\x12\x31\n\nstore_type\x18\x02 \x01(\x0e\x32\x1d.featurestorecommon.StoreType\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x1a\n\x12\x63reation_timestamp\x18\x04 \x01(\x03\x12\x1e\n\x16last_updated_timestamp\x18\x05 \x01(\x03\x12\x12\n\ncreator_id\x18\x06 \x01(\t\x12\x10\n\x04host\x18\x07 \x01(\tB\x02\x18\x01\x12\x10\n\x04port\x18\x08 \x01(\x05\x42\x02\x18\x01\x12\x1b\n\x13last_update_user_id\x18\t \x01(\t\x12;\n\x0emysql_metadata\x18\n \x01(\x0b\x32!.featurestorecommon.MySqlMetadataH\x00\x12\x44\n\x13sql_server_metadata\x18\x0b \x01(\x0b\x32%.featurestorecommon.SqlServerMetadataH\x00\x12\x41\n\x11\x64ynamodb_metadata\x18\x0c \x01(\x0b\x32$.featurestorecommon.DynamoDbMetadataH\x00\x12\x41\n\x11\x63osmosdb_metadata\x18\r \x01(\x0b\x32$.featurestorecommon.CosmosDbMetadataH\x00\x42\x15\n\x13\x61\x64\x64itional_metadata\"\xdf\x04\n\x13OnlineStoreDetailed\x12(\n\x05\x63loud\x18\x01 \x01(\x0e\x32\x19.featurestorecommon.Cloud\x12\x31\n\nstore_type\x18\x02 \x01(\x0e\x32\x1d.featurestorecommon.StoreType\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x1a\n\x12\x63reation_timestamp\x18\x04 \x01(\x03\x12\x1e\n\x16last_updated_timestamp\x18\x05 \x01(\x03\x12\x12\n\ncreator_id\x18\x06 \x01(\t\x12\x10\n\x04host\x18\x07 \x01(\tB\x02\x18\x01\x12\x10\n\x04port\x18\x08 \x01(\x05\x42\x02\x18\x01\x12\x1b\n\x13last_update_user_id\x18\t \x01(\t\x12\x10\n\x08\x66\x65\x61tures\x18\n \x03(\t\x12\x1a\n\x12read_secret_prefix\x18\x0b \x01(\t\x12;\n\x0emysql_metadata\x18\x0c \x01(\x0b\x32!.featurestorecommon.MySqlMetadataH\x00\x12\x44\n\x13sql_server_metadata\x18\r \x01(\x0b\x32%.featurestorecommon.SqlServerMetadataH\x00\x12\x41\n\x11\x64ynamodb_metadata\x18\x0e \x01(\x0b\x32$.featurestorecommon.DynamoDbMetadataH\x00\x12\x41\n\x11\x63osmosdb_metadata\x18\x0f \x01(\x0b\x32$.featurestorecommon.CosmosDbMetadataH\x00\x42\x15\n\x13\x61\x64\x64itional_metadata\"t\n FeatureTablesForSageMakerServing\x12P\n\ronline_tables\x18\x01 \x03(\x0b\x32\x39.featurestorecommon.OnlineFeatureTableForSageMakerServing\"\xfe\x02\n%OnlineFeatureTableForSageMakerServing\x12\x1a\n\x12\x66\x65\x61ture_table_name\x18\x01 \x01(\t\x12!\n\x19online_feature_table_name\x18\x02 \x01(\t\x12H\n\x0conline_store\x18\x03 \x01(\x0b\x32\x32.featurestorecommon.OnlineStoreForSageMakerServing\x12;\n\x0cprimary_keys\x18\x04 \x03(\x0b\x32%.featurestorecommon.PrimaryKeyDetails\x12\x18\n\x10\x66\x65\x61ture_table_id\x18\x05 \x01(\t\x12\x34\n\x08\x66\x65\x61tures\x18\x06 \x03(\x0b\x32\".featurestorecommon.FeatureDetails\x12?\n\x0etimestamp_keys\x18\x07 \x03(\x0b\x32\'.featurestorecommon.TimestampKeyDetails\"\xbe\x01\n\x1eOnlineStoreForSageMakerServing\x12\x1d\n\x15\x63reation_timestamp_ms\x18\x01 \x01(\x03\x12\x39\n\rdynamodb_conf\x18\x02 \x01(\x0b\x32 .featurestorecommon.DynamoDbConfH\x00\x12\x31\n\nquery_mode\x18\x10 \x01(\x0e\x32\x1d.featurestorecommon.QueryModeB\x0f\n\rextra_configs\"\x83\x01\n\x1c\x46\x65\x61tureFunctionParameterInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x11\n\ttype_text\x18\x02 \x01(\t\x12\x11\n\ttype_json\x18\x03 \x01(\t\x12/\n\ttype_name\x18\x04 \x01(\x0e\x32\x1c.featurestorecommon.DataType\"\xdb\x01\n\x19\x46\x65\x61tureFunctionForServing\x12\x11\n\tfull_name\x18\x01 \x01(\t\x12\x46\n\x0cinput_params\x18\x02 \x03(\x0b\x32\x30.featurestorecommon.FeatureFunctionParameterInfo\x12/\n\tdata_type\x18\x03 \x01(\x0e\x32\x1c.featurestorecommon.DataType\x12\x16\n\x0e\x66ull_data_type\x18\x04 \x01(\t\x12\x1a\n\x12routine_definition\x18\x05 \x01(\t\"c\n\x1a\x46\x65\x61tureFunctionsForServing\x12\x45\n\x0e\x66\x66_for_serving\x18\x01 \x03(\x0b\x32-.featurestorecommon.FeatureFunctionForServing*$\n\x05\x43loud\x12\x07\n\x03\x41WS\x10\x01\x12\t\n\x05\x41ZURE\x10\x02\x12\x07\n\x03GCP\x10\x03*T\n\tStoreType\x12\x10\n\x0c\x41URORA_MYSQL\x10\x01\x12\x0e\n\nSQL_SERVER\x10\x02\x12\t\n\x05MYSQL\x10\x03\x12\x0c\n\x08\x44YNAMODB\x10\x04\x12\x0c\n\x08\x43OSMOSDB\x10\x05*\xa2\x01\n\x08\x44\x61taType\x12\x0b\n\x07INTEGER\x10\x01\x12\t\n\x05\x46LOAT\x10\x02\x12\x0b\n\x07\x42OOLEAN\x10\x03\x12\n\n\x06STRING\x10\x04\x12\n\n\x06\x44OUBLE\x10\x05\x12\x08\n\x04LONG\x10\x06\x12\r\n\tTIMESTAMP\x10\x07\x12\x08\n\x04\x44\x41TE\x10\x08\x12\t\n\x05SHORT\x10\t\x12\t\n\x05\x41RRAY\x10\n\x12\x07\n\x03MAP\x10\x0b\x12\n\n\x06\x42INARY\x10\x0c\x12\x0b\n\x07\x44\x45\x43IMAL\x10\r*4\n\tQueryMode\x12\x16\n\x12PRIMARY_KEY_LOOKUP\x10\x00\x12\x0f\n\x0bRANGE_QUERY\x10\x01\x42\x31\n\'com.databricks.proto.featurestorecommon\xa0\x01\x01\xe2?\x02\x10\x01')

_CLOUD = DESCRIPTOR.enum_types_by_name['Cloud']
Cloud = enum_type_wrapper.EnumTypeWrapper(_CLOUD)
_STORETYPE = DESCRIPTOR.enum_types_by_name['StoreType']
StoreType = enum_type_wrapper.EnumTypeWrapper(_STORETYPE)
_DATATYPE = DESCRIPTOR.enum_types_by_name['DataType']
DataType = enum_type_wrapper.EnumTypeWrapper(_DATATYPE)
_QUERYMODE = DESCRIPTOR.enum_types_by_name['QueryMode']
QueryMode = enum_type_wrapper.EnumTypeWrapper(_QUERYMODE)
AWS = 1
AZURE = 2
GCP = 3
AURORA_MYSQL = 1
SQL_SERVER = 2
MYSQL = 3
DYNAMODB = 4
COSMOSDB = 5
INTEGER = 1
FLOAT = 2
BOOLEAN = 3
STRING = 4
DOUBLE = 5
LONG = 6
TIMESTAMP = 7
DATE = 8
SHORT = 9
ARRAY = 10
MAP = 11
BINARY = 12
DECIMAL = 13
PRIMARY_KEY_LOOKUP = 0
RANGE_QUERY = 1


_ONLINESTOREFORSERVING = DESCRIPTOR.message_types_by_name['OnlineStoreForServing']
_MYSQLCONF = DESCRIPTOR.message_types_by_name['MySqlConf']
_SQLSERVERCONF = DESCRIPTOR.message_types_by_name['SqlServerConf']
_DYNAMODBCONF = DESCRIPTOR.message_types_by_name['DynamoDbConf']
_COSMOSDBCONF = DESCRIPTOR.message_types_by_name['CosmosDbConf']
_FEATUREDETAILS = DESCRIPTOR.message_types_by_name['FeatureDetails']
_PRIMARYKEYDETAILS = DESCRIPTOR.message_types_by_name['PrimaryKeyDetails']
_TIMESTAMPKEYDETAILS = DESCRIPTOR.message_types_by_name['TimestampKeyDetails']
_ONLINEFEATURETABLE = DESCRIPTOR.message_types_by_name['OnlineFeatureTable']
_FEATURETABLESFORSERVING = DESCRIPTOR.message_types_by_name['FeatureTablesForServing']
_MYSQLMETADATA = DESCRIPTOR.message_types_by_name['MySqlMetadata']
_SQLSERVERMETADATA = DESCRIPTOR.message_types_by_name['SqlServerMetadata']
_DYNAMODBMETADATA = DESCRIPTOR.message_types_by_name['DynamoDbMetadata']
_COSMOSDBMETADATA = DESCRIPTOR.message_types_by_name['CosmosDbMetadata']
_ONLINESTORE = DESCRIPTOR.message_types_by_name['OnlineStore']
_ONLINESTOREDETAILED = DESCRIPTOR.message_types_by_name['OnlineStoreDetailed']
_FEATURETABLESFORSAGEMAKERSERVING = DESCRIPTOR.message_types_by_name['FeatureTablesForSageMakerServing']
_ONLINEFEATURETABLEFORSAGEMAKERSERVING = DESCRIPTOR.message_types_by_name['OnlineFeatureTableForSageMakerServing']
_ONLINESTOREFORSAGEMAKERSERVING = DESCRIPTOR.message_types_by_name['OnlineStoreForSageMakerServing']
_FEATUREFUNCTIONPARAMETERINFO = DESCRIPTOR.message_types_by_name['FeatureFunctionParameterInfo']
_FEATUREFUNCTIONFORSERVING = DESCRIPTOR.message_types_by_name['FeatureFunctionForServing']
_FEATUREFUNCTIONSFORSERVING = DESCRIPTOR.message_types_by_name['FeatureFunctionsForServing']
OnlineStoreForServing = _reflection.GeneratedProtocolMessageType('OnlineStoreForServing', (_message.Message,), {
  'DESCRIPTOR' : _ONLINESTOREFORSERVING,
  '__module__' : 'feature_store_serving_pb2'
  # @@protoc_insertion_point(class_scope:featurestorecommon.OnlineStoreForServing)
  })
_sym_db.RegisterMessage(OnlineStoreForServing)

MySqlConf = _reflection.GeneratedProtocolMessageType('MySqlConf', (_message.Message,), {
  'DESCRIPTOR' : _MYSQLCONF,
  '__module__' : 'feature_store_serving_pb2'
  # @@protoc_insertion_point(class_scope:featurestorecommon.MySqlConf)
  })
_sym_db.RegisterMessage(MySqlConf)

SqlServerConf = _reflection.GeneratedProtocolMessageType('SqlServerConf', (_message.Message,), {
  'DESCRIPTOR' : _SQLSERVERCONF,
  '__module__' : 'feature_store_serving_pb2'
  # @@protoc_insertion_point(class_scope:featurestorecommon.SqlServerConf)
  })
_sym_db.RegisterMessage(SqlServerConf)

DynamoDbConf = _reflection.GeneratedProtocolMessageType('DynamoDbConf', (_message.Message,), {
  'DESCRIPTOR' : _DYNAMODBCONF,
  '__module__' : 'feature_store_serving_pb2'
  # @@protoc_insertion_point(class_scope:featurestorecommon.DynamoDbConf)
  })
_sym_db.RegisterMessage(DynamoDbConf)

CosmosDbConf = _reflection.GeneratedProtocolMessageType('CosmosDbConf', (_message.Message,), {
  'DESCRIPTOR' : _COSMOSDBCONF,
  '__module__' : 'feature_store_serving_pb2'
  # @@protoc_insertion_point(class_scope:featurestorecommon.CosmosDbConf)
  })
_sym_db.RegisterMessage(CosmosDbConf)

FeatureDetails = _reflection.GeneratedProtocolMessageType('FeatureDetails', (_message.Message,), {
  'DESCRIPTOR' : _FEATUREDETAILS,
  '__module__' : 'feature_store_serving_pb2'
  # @@protoc_insertion_point(class_scope:featurestorecommon.FeatureDetails)
  })
_sym_db.RegisterMessage(FeatureDetails)

PrimaryKeyDetails = _reflection.GeneratedProtocolMessageType('PrimaryKeyDetails', (_message.Message,), {
  'DESCRIPTOR' : _PRIMARYKEYDETAILS,
  '__module__' : 'feature_store_serving_pb2'
  # @@protoc_insertion_point(class_scope:featurestorecommon.PrimaryKeyDetails)
  })
_sym_db.RegisterMessage(PrimaryKeyDetails)

TimestampKeyDetails = _reflection.GeneratedProtocolMessageType('TimestampKeyDetails', (_message.Message,), {
  'DESCRIPTOR' : _TIMESTAMPKEYDETAILS,
  '__module__' : 'feature_store_serving_pb2'
  # @@protoc_insertion_point(class_scope:featurestorecommon.TimestampKeyDetails)
  })
_sym_db.RegisterMessage(TimestampKeyDetails)

OnlineFeatureTable = _reflection.GeneratedProtocolMessageType('OnlineFeatureTable', (_message.Message,), {
  'DESCRIPTOR' : _ONLINEFEATURETABLE,
  '__module__' : 'feature_store_serving_pb2'
  # @@protoc_insertion_point(class_scope:featurestorecommon.OnlineFeatureTable)
  })
_sym_db.RegisterMessage(OnlineFeatureTable)

FeatureTablesForServing = _reflection.GeneratedProtocolMessageType('FeatureTablesForServing', (_message.Message,), {
  'DESCRIPTOR' : _FEATURETABLESFORSERVING,
  '__module__' : 'feature_store_serving_pb2'
  # @@protoc_insertion_point(class_scope:featurestorecommon.FeatureTablesForServing)
  })
_sym_db.RegisterMessage(FeatureTablesForServing)

MySqlMetadata = _reflection.GeneratedProtocolMessageType('MySqlMetadata', (_message.Message,), {
  'DESCRIPTOR' : _MYSQLMETADATA,
  '__module__' : 'feature_store_serving_pb2'
  # @@protoc_insertion_point(class_scope:featurestorecommon.MySqlMetadata)
  })
_sym_db.RegisterMessage(MySqlMetadata)

SqlServerMetadata = _reflection.GeneratedProtocolMessageType('SqlServerMetadata', (_message.Message,), {
  'DESCRIPTOR' : _SQLSERVERMETADATA,
  '__module__' : 'feature_store_serving_pb2'
  # @@protoc_insertion_point(class_scope:featurestorecommon.SqlServerMetadata)
  })
_sym_db.RegisterMessage(SqlServerMetadata)

DynamoDbMetadata = _reflection.GeneratedProtocolMessageType('DynamoDbMetadata', (_message.Message,), {
  'DESCRIPTOR' : _DYNAMODBMETADATA,
  '__module__' : 'feature_store_serving_pb2'
  # @@protoc_insertion_point(class_scope:featurestorecommon.DynamoDbMetadata)
  })
_sym_db.RegisterMessage(DynamoDbMetadata)

CosmosDbMetadata = _reflection.GeneratedProtocolMessageType('CosmosDbMetadata', (_message.Message,), {
  'DESCRIPTOR' : _COSMOSDBMETADATA,
  '__module__' : 'feature_store_serving_pb2'
  # @@protoc_insertion_point(class_scope:featurestorecommon.CosmosDbMetadata)
  })
_sym_db.RegisterMessage(CosmosDbMetadata)

OnlineStore = _reflection.GeneratedProtocolMessageType('OnlineStore', (_message.Message,), {
  'DESCRIPTOR' : _ONLINESTORE,
  '__module__' : 'feature_store_serving_pb2'
  # @@protoc_insertion_point(class_scope:featurestorecommon.OnlineStore)
  })
_sym_db.RegisterMessage(OnlineStore)

OnlineStoreDetailed = _reflection.GeneratedProtocolMessageType('OnlineStoreDetailed', (_message.Message,), {
  'DESCRIPTOR' : _ONLINESTOREDETAILED,
  '__module__' : 'feature_store_serving_pb2'
  # @@protoc_insertion_point(class_scope:featurestorecommon.OnlineStoreDetailed)
  })
_sym_db.RegisterMessage(OnlineStoreDetailed)

FeatureTablesForSageMakerServing = _reflection.GeneratedProtocolMessageType('FeatureTablesForSageMakerServing', (_message.Message,), {
  'DESCRIPTOR' : _FEATURETABLESFORSAGEMAKERSERVING,
  '__module__' : 'feature_store_serving_pb2'
  # @@protoc_insertion_point(class_scope:featurestorecommon.FeatureTablesForSageMakerServing)
  })
_sym_db.RegisterMessage(FeatureTablesForSageMakerServing)

OnlineFeatureTableForSageMakerServing = _reflection.GeneratedProtocolMessageType('OnlineFeatureTableForSageMakerServing', (_message.Message,), {
  'DESCRIPTOR' : _ONLINEFEATURETABLEFORSAGEMAKERSERVING,
  '__module__' : 'feature_store_serving_pb2'
  # @@protoc_insertion_point(class_scope:featurestorecommon.OnlineFeatureTableForSageMakerServing)
  })
_sym_db.RegisterMessage(OnlineFeatureTableForSageMakerServing)

OnlineStoreForSageMakerServing = _reflection.GeneratedProtocolMessageType('OnlineStoreForSageMakerServing', (_message.Message,), {
  'DESCRIPTOR' : _ONLINESTOREFORSAGEMAKERSERVING,
  '__module__' : 'feature_store_serving_pb2'
  # @@protoc_insertion_point(class_scope:featurestorecommon.OnlineStoreForSageMakerServing)
  })
_sym_db.RegisterMessage(OnlineStoreForSageMakerServing)

FeatureFunctionParameterInfo = _reflection.GeneratedProtocolMessageType('FeatureFunctionParameterInfo', (_message.Message,), {
  'DESCRIPTOR' : _FEATUREFUNCTIONPARAMETERINFO,
  '__module__' : 'feature_store_serving_pb2'
  # @@protoc_insertion_point(class_scope:featurestorecommon.FeatureFunctionParameterInfo)
  })
_sym_db.RegisterMessage(FeatureFunctionParameterInfo)

FeatureFunctionForServing = _reflection.GeneratedProtocolMessageType('FeatureFunctionForServing', (_message.Message,), {
  'DESCRIPTOR' : _FEATUREFUNCTIONFORSERVING,
  '__module__' : 'feature_store_serving_pb2'
  # @@protoc_insertion_point(class_scope:featurestorecommon.FeatureFunctionForServing)
  })
_sym_db.RegisterMessage(FeatureFunctionForServing)

FeatureFunctionsForServing = _reflection.GeneratedProtocolMessageType('FeatureFunctionsForServing', (_message.Message,), {
  'DESCRIPTOR' : _FEATUREFUNCTIONSFORSERVING,
  '__module__' : 'feature_store_serving_pb2'
  # @@protoc_insertion_point(class_scope:featurestorecommon.FeatureFunctionsForServing)
  })
_sym_db.RegisterMessage(FeatureFunctionsForServing)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\'com.databricks.proto.featurestorecommon\240\001\001\342?\002\020\001'
  _ONLINESTORE.fields_by_name['host']._options = None
  _ONLINESTORE.fields_by_name['host']._serialized_options = b'\030\001'
  _ONLINESTORE.fields_by_name['port']._options = None
  _ONLINESTORE.fields_by_name['port']._serialized_options = b'\030\001'
  _ONLINESTOREDETAILED.fields_by_name['host']._options = None
  _ONLINESTOREDETAILED.fields_by_name['host']._serialized_options = b'\030\001'
  _ONLINESTOREDETAILED.fields_by_name['port']._options = None
  _ONLINESTOREDETAILED.fields_by_name['port']._serialized_options = b'\030\001'
  _CLOUD._serialized_start=3991
  _CLOUD._serialized_end=4027
  _STORETYPE._serialized_start=4029
  _STORETYPE._serialized_end=4113
  _DATATYPE._serialized_start=4116
  _DATATYPE._serialized_end=4278
  _QUERYMODE._serialized_start=4280
  _QUERYMODE._serialized_end=4332
  _ONLINESTOREFORSERVING._serialized_start=75
  _ONLINESTOREFORSERVING._serialized_end=551
  _MYSQLCONF._serialized_start=553
  _MYSQLCONF._serialized_end=592
  _SQLSERVERCONF._serialized_start=594
  _SQLSERVERCONF._serialized_end=637
  _DYNAMODBCONF._serialized_start=639
  _DYNAMODBCONF._serialized_end=669
  _COSMOSDBCONF._serialized_start=671
  _COSMOSDBCONF._serialized_end=706
  _FEATUREDETAILS._serialized_start=708
  _FEATUREDETAILS._serialized_end=814
  _PRIMARYKEYDETAILS._serialized_start=816
  _PRIMARYKEYDETAILS._serialized_end=898
  _TIMESTAMPKEYDETAILS._serialized_start=900
  _TIMESTAMPKEYDETAILS._serialized_end=984
  _ONLINEFEATURETABLE._serialized_start=987
  _ONLINEFEATURETABLE._serialized_end=1341
  _FEATURETABLESFORSERVING._serialized_start=1343
  _FEATURETABLESFORSERVING._serialized_end=1431
  _MYSQLMETADATA._serialized_start=1433
  _MYSQLMETADATA._serialized_end=1476
  _SQLSERVERMETADATA._serialized_start=1478
  _SQLSERVERMETADATA._serialized_end=1525
  _DYNAMODBMETADATA._serialized_start=1527
  _DYNAMODBMETADATA._serialized_end=1593
  _COSMOSDBMETADATA._serialized_start=1595
  _COSMOSDBMETADATA._serialized_end=1670
  _ONLINESTORE._serialized_start=1673
  _ONLINESTORE._serialized_end=2226
  _ONLINESTOREDETAILED._serialized_start=2229
  _ONLINESTOREDETAILED._serialized_end=2836
  _FEATURETABLESFORSAGEMAKERSERVING._serialized_start=2838
  _FEATURETABLESFORSAGEMAKERSERVING._serialized_end=2954
  _ONLINEFEATURETABLEFORSAGEMAKERSERVING._serialized_start=2957
  _ONLINEFEATURETABLEFORSAGEMAKERSERVING._serialized_end=3339
  _ONLINESTOREFORSAGEMAKERSERVING._serialized_start=3342
  _ONLINESTOREFORSAGEMAKERSERVING._serialized_end=3532
  _FEATUREFUNCTIONPARAMETERINFO._serialized_start=3535
  _FEATUREFUNCTIONPARAMETERINFO._serialized_end=3666
  _FEATUREFUNCTIONFORSERVING._serialized_start=3669
  _FEATUREFUNCTIONFORSERVING._serialized_end=3888
  _FEATUREFUNCTIONSFORSERVING._serialized_start=3890
  _FEATUREFUNCTIONSFORSERVING._serialized_end=3989
# @@protoc_insertion_point(module_scope)



import copy
import re
from typing import List, Set

from databricks.feature_store.entities.feature_spec import FeatureSpec

SINGLE_LEVEL_NAMESPACE_TABLE_REGEX = r"^[\w_]+$"
TWO_LEVEL_NAMESPACE_TABLE_REGEX = r"^[\w_]+(\.[\w_]+)$"
THREE_LEVEL_NAMESPACE_TABLE_REGEX = r"^[\w_]+(\.[\w_]+){2}$"

HIVE_METASTORE_NAME = "hive_metastore"
# these two catalog names both points to the workspace local default HMS (hive metastore).
LOCAL_METASTORE_NAMES = [HIVE_METASTORE_NAME, "spark_catalog"]
# samples catalog is managed by databricks for hosting public dataset like NYC taxi dataset.
# it is neither a UC nor local metastore catalog, accessing samples catalog using feature store client is not allowed.
SAMPLES_CATALOG_NAME = "samples"


# Get full table name in the form of <catalog_name>.<database_name>.<table_name>
# given user specified table name, current catalog and schema.
def get_full_table_name(
    table_name: str,
    current_catalog: str,
    current_schema: str,
) -> str:
    if not _is_single_level_name(current_catalog) or not _is_single_level_name(
        current_schema
    ):
        raise ValueError(
            f"Invalid catalog '{current_catalog}' or "
            f"schema '{current_schema}' name for table '{table_name}'."
        )
    _check_qualified_names({table_name})
    if _is_single_level_name(table_name):
        full_table_name = f"{current_catalog}.{current_schema}.{table_name}"
    elif _is_two_level_name(table_name):
        full_table_name = f"{current_catalog}.{table_name}"
    elif _is_three_level_name(table_name):
        full_table_name = table_name
    else:
        raise _invalid_table_names_error({table_name})
    catalog, schema, table = full_table_name.split(".")
    if catalog == SAMPLES_CATALOG_NAME:
        raise ValueError(
            "'samples' catalog cannot be accessed using feature store client."
        )
    if catalog in LOCAL_METASTORE_NAMES:
        return f"{HIVE_METASTORE_NAME}.{schema}.{table}"
    return full_table_name


# Local metastore tables in feature_spec.yaml are all stored in 2L.
# Standardize table names to be all in 3L to avoid erroneously reading data from UC tables.
def get_feature_spec_with_full_table_names(feature_spec: FeatureSpec) -> FeatureSpec:
    column_info_table_names = [
        column_info.table_name for column_info in feature_spec.feature_column_infos
    ]
    table_info_table_names = [
        table_info.table_name for table_info in feature_spec.table_infos
    ]
    _check_qualified_names(set(column_info_table_names))
    _check_qualified_names(set(table_info_table_names))
    invalid_table_names = list(
        filter(_is_single_level_name, column_info_table_names)
    ) + list(filter(_is_single_level_name, table_info_table_names))
    if len(invalid_table_names) > 0:
        raise _invalid_table_names_error(set(invalid_table_names))
    standardized_feature_spec = copy.deepcopy(feature_spec)
    for column_info in standardized_feature_spec.feature_column_infos:
        if _is_two_level_name(column_info.table_name):
            column_info._table_name = f"{HIVE_METASTORE_NAME}.{column_info.table_name}"
    for table_info in standardized_feature_spec.table_infos:
        if _is_two_level_name(table_info.table_name):
            table_info._table_name = f"{HIVE_METASTORE_NAME}.{table_info.table_name}"
    return standardized_feature_spec


# Reformat 3L table name for tables in local metastore to 2L. This is used when interacting with catalog client
# and serializing workspace local feature spec for scoring.
def reformat_full_table_name(full_table_name: str) -> str:
    if not _is_three_level_name(full_table_name):
        raise _invalid_table_names_error({full_table_name})
    catalog, schema, table = full_table_name.split(".")
    if catalog in LOCAL_METASTORE_NAMES:
        return f"{schema}.{table}"
    return full_table_name


# Reformat table names in feature_spec with reformat_full_table_name
def get_feature_spec_with_reformat_full_table_names(
    feature_spec: FeatureSpec,
) -> FeatureSpec:
    column_info_table_names = [
        column_info.table_name for column_info in feature_spec.feature_column_infos
    ]
    table_info_table_names = [
        table_info.table_name for table_info in feature_spec.table_infos
    ]
    _check_qualified_names(set(column_info_table_names))
    _check_qualified_names(set(table_info_table_names))
    invalid_table_names = list(
        filter(lambda name: not _is_three_level_name(name), column_info_table_names)
    ) + list(
        filter(lambda name: not _is_three_level_name(name), table_info_table_names)
    )
    if len(invalid_table_names) > 0:
        raise _invalid_table_names_error(set(invalid_table_names))
    standardized_feature_spec = copy.deepcopy(feature_spec)
    for column_info in standardized_feature_spec.feature_column_infos:
        column_info._table_name = reformat_full_table_name(column_info.table_name)
    for table_info in standardized_feature_spec.table_infos:
        table_info._table_name = reformat_full_table_name(table_info.table_name)
    return standardized_feature_spec


def _invalid_table_names_error(invalid_table_names: Set[str]) -> ValueError:
    return ValueError(
        f"Invalid table name{'s' if len(invalid_table_names) > 1 else ''} '{', '.join(invalid_table_names)}'."
    )


def _is_qualified_table_name(feature_table_name) -> bool:
    return isinstance(feature_table_name, str) and (
        _is_single_level_name(feature_table_name)
        or _is_two_level_name(feature_table_name)
        or _is_three_level_name(feature_table_name)
    )


def _is_single_level_name(name) -> bool:
    return (
        isinstance(name, str)
        and re.match(SINGLE_LEVEL_NAMESPACE_TABLE_REGEX, name) is not None
    )


def _is_two_level_name(name) -> bool:
    return (
        isinstance(name, str)
        and re.match(TWO_LEVEL_NAMESPACE_TABLE_REGEX, name) is not None
    )


def _is_three_level_name(name) -> bool:
    return (
        isinstance(name, str)
        and re.match(THREE_LEVEL_NAMESPACE_TABLE_REGEX, name) is not None
    )


def tagging_for_uc_tables_error(full_table_name):
    return ValueError(
        f"Cannot set or delete tags for table {full_table_name}. "
        f"Tagging for Feature Tables in Unity Catalog is not yet supported."
    )


# check if table is in UC
def is_uc_table(full_table_name) -> bool:
    catalog_name, schema_name, table_name = full_table_name.split(".")
    return (
        not is_default_hms_table(full_table_name)
        and catalog_name != SAMPLES_CATALOG_NAME
    )


def is_default_hms_table(full_table_name) -> bool:
    catalog_name, schema_name, table_name = full_table_name.split(".")
    return (
        catalog_name in LOCAL_METASTORE_NAMES and catalog_name != SAMPLES_CATALOG_NAME
    )


# check if table names are in the correct format - 1L, 2L or 3L
def _check_qualified_names(feature_table_names: Set[str]):
    unqualified_table_names = list(
        filter(
            lambda table_name: not _is_qualified_table_name(table_name),
            feature_table_names,
        )
    )
    if len(unqualified_table_names) > 0:
        raise ValueError(
            f"Feature table name{'s' if len(unqualified_table_names) > 1 else ''} "
            f"'{', '.join(map(str, unqualified_table_names))}' must have the form "
            f"<catalog_name>.<schema_name>.<table_name>, <database_name>.<table_name>, "
            f"or <table_name> and only contain alphabet characters, numbers and _."
        )


# For APIs like create_training_set and score_batch, all tables must all be in
# UC catalog (shareable cross-workspaces) or default HMS (intended to only be used in the current workspace)
# check if all tables are either in UC or default HMS.
def _verify_all_tables_are_either_in_uc_or_in_hms(
    table_names: Set[str], current_catalog: str, current_schema: str
):
    is_valid = all(
        map(
            lambda table_name: is_uc_table(
                get_full_table_name(table_name, current_catalog, current_schema)
            ),
            table_names,
        )
    ) or all(
        map(
            lambda table_name: is_default_hms_table(
                get_full_table_name(table_name, current_catalog, current_schema)
            ),
            table_names,
        )
    )
    if not is_valid:
        raise ValueError(
            f"Feature table names '{', '.join(table_names)}' "
            f"must all be in UC or the local default hive metastore. "
            f"Mixing feature tables from two different storage locations is not allowed."
        )



from databricks.feature_store.entities._feature_store_object import _FeatureStoreObject
from databricks.feature_store.protos.feature_spec_pb2 import (
    FeatureTableInfo as ProtoFeatureTableInfo,
)


class FeatureTableInfo(_FeatureStoreObject):
    def __init__(self, table_name: str, table_id: str):
        if not table_name:
            raise ValueError("table_name must be non-empty.")
        if not table_id:
            raise ValueError("table_id must be non-empty.")
        self._table_name = table_name
        self._table_id = table_id

    @property
    def table_name(self):
        return self._table_name

    @property
    def table_id(self):
        return self._table_id

    @classmethod
    def from_proto(cls, feature_table_info_proto):
        return cls(
            table_name=feature_table_info_proto.table_name,
            table_id=feature_table_info_proto.table_id,
        )

    def to_proto(self):
        return ProtoFeatureTableInfo(
            table_name=self.table_name,
            table_id=self.table_id,
        )



from typing import Any

from databricks.feature_store.entities._proto_enum_entity import _ProtoEnumEntity
from databricks.feature_store.protos.feature_store_serving_pb2 import (
    QueryMode as ProtoQueryMode,
)


class QueryMode(_ProtoEnumEntity):
    """Online store query modes."""

    PRIMARY_KEY_LOOKUP = ProtoQueryMode.Value("PRIMARY_KEY_LOOKUP")
    RANGE_QUERY = ProtoQueryMode.Value("RANGE_QUERY")

    @classmethod
    def _enum_type(cls) -> Any:
        return ProtoQueryMode



from typing import Any

from databricks.feature_store.entities._proto_enum_entity import _ProtoEnumEntity
from databricks.feature_store.protos.feature_store_serving_pb2 import (
    StoreType as ProtoStoreType,
)


class StoreType(_ProtoEnumEntity):
    """Online store types."""

    AURORA_MYSQL = ProtoStoreType.Value("AURORA_MYSQL")
    SQL_SERVER = ProtoStoreType.Value("SQL_SERVER")
    MYSQL = ProtoStoreType.Value("MYSQL")
    DYNAMODB = ProtoStoreType.Value("DYNAMODB")
    COSMOSDB = ProtoStoreType.Value("COSMOSDB")

    @classmethod
    def _enum_type(cls) -> Any:
        return ProtoStoreType



import abc


class _ProtoEnumEntity(object):
    """
    Generic entity to map to proto enum messages.
    When inheriting from this class override `_enum_type` method to return underlying enum proto.
    """

    _STRING_TO_ENUM, _ENUM_TO_STRING = None, None

    @classmethod
    def init(cls):
        if not cls._STRING_TO_ENUM:
            enum_proto = cls._enum_type()
            cls._STRING_TO_ENUM = {k: enum_proto.Value(k) for k in enum_proto.keys()}
            cls._ENUM_TO_STRING = {
                value: key for key, value in cls._STRING_TO_ENUM.items()
            }

    @classmethod
    @abc.abstractmethod
    def _enum_type(cls) -> str:
        pass

    @classmethod
    def from_string(cls, str_value):
        cls.init()
        if str_value.upper() not in cls._STRING_TO_ENUM:
            raise Exception(
                f"Could not find a {cls._enum_type().DESCRIPTOR.name} value corresponding to the "
                f" input string '{str_value}'. Valid values: '{list(cls._STRING_TO_ENUM.keys())}'"
            )
        return cls._STRING_TO_ENUM[str_value.upper()]

    @classmethod
    def to_string(cls, value):
        cls.init()
        if value not in cls._ENUM_TO_STRING:
            raise Exception(
                f"Input {cls._enum_type().DESCRIPTOR.name} type '{value}' cannot be converted to"
                f" a string representation. Valid inputs: '{list(cls._ENUM_TO_STRING.keys())}'"
            )
        return cls._ENUM_TO_STRING[value]



import collections

import sqlalchemy

from databricks.feature_store.entities.cloud import Cloud
from databricks.feature_store.entities.online_feature_table import OnlineFeatureTable
from databricks.feature_store.lookup_engine.lookup_sql_engine import LookupSqlEngine


class LookupMySqlEngine(LookupSqlEngine):

    PRIMARY_COLUMN_KEY = "PRI"

    def __init__(
        self, online_feature_table: OnlineFeatureTable, ro_user: str, ro_password: str
    ):
        super().__init__(online_feature_table, ro_user, ro_password)

    @property
    def engine_url(self):
        if self.online_store.cloud == Cloud.AZURE:
            """
            Construct an engine URL in the format that Azure mysql expects, where the user portion
            of the connection string is in the <username@hostname> format.  The rationale for this
            format is described in the Azure docs here: https://bit.ly/3BGBBFF.

            This function expects self.host to be in the format: <servername>.mysql.database.azure.com, which is the expected
            format according to the Azure docs mentioned above.
            """
            if "." in self.host:
                hostname = self.host.split(".")[0]
            else:
                raise ValueError(
                    "AzureMySql host not in the expected format: <servername>.mysql.database.azure.com"
                )
            username = f"{self.user}@{hostname}"
        else:
            username = self.user

        return f"mysql+pymysql://{username}:{self.password}@{self.host}:{self.port}/{self.database_name}"

    @classmethod
    def _sql_safe_name(cls, name):
        # MySQL requires `xxx` format to safely handle identifiers that contain special characters or are reserved words.
        return f"`{name}`"

    def _database_contains_feature_table(self, sql_connection):
        query = sqlalchemy.sql.text(
            f"SELECT {self.TABLE_NAME} FROM {self.INFORMATION_SCHEMA}.{self.TABLES} "
            f"WHERE {self.TABLE_SCHEMA}='{self.database_name}' AND {self.TABLE_NAME} IN ('{self.table_name}')"
        )
        results = sql_connection.execute(query)
        table = results.fetchall()
        return len(table) > 0

    def _database_contains_primary_keys(self, sql_connection):
        query = sqlalchemy.sql.text(
            f"SELECT {self.COLUMN_NAME} FROM {self.INFORMATION_SCHEMA}.{self.COLUMNS} "
            f"WHERE {self.TABLE_SCHEMA}='{self.database_name}' AND {self.TABLE_NAME}='{self.table_name}' "
            f"AND {self.COLUMN_KEY}='{self.PRIMARY_COLUMN_KEY}'"
        )
        results = sql_connection.execute(query)
        primary_keys = [r[0] for r in results.fetchall()]
        return collections.Counter(primary_keys) == collections.Counter(
            [primary_key.name for primary_key in self.primary_keys]
        )



from typing import Union

from databricks.feature_store.entities._feature_store_object import _FeatureStoreObject
from databricks.feature_store.entities.cloud import Cloud
from databricks.feature_store.entities.query_mode import QueryMode
from databricks.feature_store.entities.store_type import StoreType
from databricks.feature_store.protos.feature_store_serving_pb2 import (
    DynamoDbConf as ProtoDynamoDbConf,
)
from databricks.feature_store.protos.feature_store_serving_pb2 import (
    OnlineStoreForSageMakerServing as ProtoOnlineStoreForSageMakerServing,
)


class MySqlConf(_FeatureStoreObject):
    def __init__(self, host: str, port: int):
        self._host = host
        self._port = port

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port


class SqlServerConf(_FeatureStoreObject):
    def __init__(self, host: str, port: int):
        self._host = host
        self._port = port

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port


class DynamoDbConf(_FeatureStoreObject):
    def __init__(self, region: str):
        self._region = region

    @property
    def region(self):
        return self._region


class CosmosDbConf(_FeatureStoreObject):
    def __init__(self, account_uri: str):
        self._account_uri = account_uri

    @property
    def account_uri(self):
        return self._account_uri


class AbstractOnlineStoreForServing(_FeatureStoreObject):
    def __init__(
        self,
        creation_timestamp_ms: int,
        extra_configs: Union[MySqlConf, SqlServerConf, DynamoDbConf],
        query_mode: QueryMode,
    ):
        self._creation_timestamp_ms = creation_timestamp_ms
        self._extra_configs = extra_configs
        self._query_mode = query_mode

    @property
    def extra_configs(self):
        return self._extra_configs

    @property
    def creation_timestamp_ms(self):
        return self._creation_timestamp_ms

    @property
    def query_mode(self) -> QueryMode:
        return self._query_mode


class OnlineStoreForServing(AbstractOnlineStoreForServing):
    def __init__(
        self,
        cloud: Cloud,
        store_type: StoreType,
        read_secret_prefix: str,
        **kwargs,
    ):
        self._cloud = cloud
        self._store_type = store_type
        self._read_secret_prefix = read_secret_prefix
        super().__init__(**kwargs)

    @property
    def cloud(self):
        return self._cloud

    @property
    def store_type(self):
        return self._store_type

    @property
    def read_secret_prefix(self):
        return self._read_secret_prefix

    @classmethod
    def from_proto(cls, the_proto):
        conf = None
        if the_proto.WhichOneof("extra_configs") == "mysql_conf":
            conf = MySqlConf(the_proto.mysql_conf.host, the_proto.mysql_conf.port)
        elif the_proto.WhichOneof("extra_configs") == "sql_server_conf":
            conf = SqlServerConf(
                the_proto.sql_server_conf.host, the_proto.sql_server_conf.port
            )
        elif the_proto.WhichOneof("extra_configs") == "dynamodb_conf":
            conf = DynamoDbConf(the_proto.dynamodb_conf.region)
        elif the_proto.WhichOneof("extra_configs") == "cosmosdb_conf":
            conf = CosmosDbConf(the_proto.cosmosdb_conf.account_uri)
        else:
            raise ValueError("Unsupported Store Type: " + str(the_proto))

        # The `query_mode` should always be defined. We explicitly check this to avoid erroneously retrieving the
        # default QueryMode enum value `PRIMARY_KEY_LOOKUP` when the `query_mode` proto field is not set.
        if not the_proto.HasField("query_mode"):
            raise ValueError(f"'query_mode' should be defined: {the_proto}")
        return cls(
            cloud=the_proto.cloud,
            store_type=the_proto.store_type,
            extra_configs=conf,
            read_secret_prefix=the_proto.read_secret_prefix,
            creation_timestamp_ms=the_proto.creation_timestamp_ms,
            query_mode=the_proto.query_mode,
        )


class OnlineStoreForSageMakerServing(AbstractOnlineStoreForServing):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def to_proto(self) -> ProtoOnlineStoreForSageMakerServing:
        return ProtoOnlineStoreForSageMakerServing(
            creation_timestamp_ms=self.creation_timestamp_ms,
            dynamodb_conf=ProtoDynamoDbConf(region=self.extra_configs.region),
            query_mode=self.query_mode,
        )

    @classmethod
    def from_proto(cls, the_proto: ProtoOnlineStoreForSageMakerServing):
        if not (the_proto.WhichOneof("extra_configs") == "dynamodb_conf"):
            raise Exception(
                "Internal Error: Store type is not supported for SageMaker: "
                f"{the_proto}"
            )
        conf = DynamoDbConf(the_proto.dynamodb_conf.region)

        # The `query_mode` should always be defined. We explicitly check this to avoid erroneously retrieving the
        # default QueryMode enum value `PRIMARY_KEY_LOOKUP` when the `query_mode` proto field is not set.
        if not the_proto.HasField("query_mode"):
            raise ValueError(f"'query_mode' should be defined: {the_proto}")
        return cls(
            creation_timestamp_ms=the_proto.creation_timestamp_ms,
            extra_configs=conf,
            query_mode=the_proto.query_mode,
        )



import itertools
import os
from collections import Counter, defaultdict
from typing import Dict, List, Tuple, Union

import mlflow
import pandas as pd
from mlflow.models.container import SERVING_ENVIRONMENT
from mlflow.sagemaker import SAGEMAKER_SERVING_ENVIRONMENT
from mlflow.utils import databricks_utils

from databricks.feature_store import mlflow_model_constants
from databricks.feature_store.entities.feature_column_info import FeatureColumnInfo
from databricks.feature_store.entities.feature_spec import FeatureSpec
from databricks.feature_store.entities.feature_tables_for_serving import (
    AbstractFeatureTablesForServing,
    FeatureTablesForSageMakerServing,
    FeatureTablesForServing,
)
from databricks.feature_store.entities.online_feature_table import (
    AbstractOnlineFeatureTable,
    OnlineFeatureTable,
    PrimaryKeyDetails,
)
from databricks.feature_store.entities.store_type import StoreType
from databricks.feature_store.feature_lookup_version import VERSION
from databricks.feature_store.online_lookup_client import (
    OnlineLookupClient,
    is_primary_key_lookup,
    tables_share_dynamodb_access_keys,
)
from databricks.feature_store.utils.feature_serving_patch import (
    patch_feature_serving_predictions_to_json,
)
from databricks.feature_store.utils.metrics_utils import (
    OVERRIDEN_FEATURE_COUNT,
    LookupClientMetrics,
    lookup_call_maybe_with_metrics,
)
from databricks.feature_store.utils.uc_utils import (
    get_feature_spec_with_full_table_names,
    get_feature_spec_with_reformat_full_table_names,
)

# The provisioner of this model is expected to set an environment variable with the path to a
# feature_tables_for_serving.dat file.
FEATURE_TABLES_FOR_SERVING_FILEPATH_ENV = "FEATURE_TABLES_FOR_SERVING_FILEPATH"
# should match LOOKUP_CLIENT_INSTRUMENTATION_ENABLED_ENV in
# model-serving/model-serving-common/src/main/scala/com/databricks/mlflow/utils/serving/FeatureStoreConstants.scala
LOOKUP_CLIENT_INSTRUMENTATION_ENABLED_ENV = "LOOKUP_CLIENT_INSTRUMENTATION_ENABLED"

FEATURE_SERVING_RESPONSE_FORMAT_ENABLED = False

LookupKeyType = Tuple[str, ...]
LookupKeyToFeatureColumnInfosType = Dict[LookupKeyType, List[FeatureColumnInfo]]


class _FeatureTableMetadata:
    """
    Encapsulates metadata on a feature table, including lookup keys, feature metadata, and online
    feature table metadata.

    Feature metadata is grouped by lookup key, since different features may require different
    lookup keys (eg pickup_zip and dropoff_zip may each be used to lookup a geographic data feature
    table).
    """

    def __init__(
        self,
        feature_col_infos_by_lookup_key: LookupKeyToFeatureColumnInfosType,
        online_ft: AbstractOnlineFeatureTable,
    ):
        self.feature_col_infos_by_lookup_key = feature_col_infos_by_lookup_key
        self.online_ft = online_ft


def _load_raw_model(path):
    raw_model_path = os.path.join(path, mlflow_model_constants.RAW_MODEL_FOLDER)
    return mlflow.pyfunc.load_model(raw_model_path)


class _FeatureStoreModelWrapper:
    def __init__(self, path: str):
        print(f"Initializing feature store lookup client: ${VERSION}")
        self._check_support()
        self.serving_environment = self._get_serving_environment()
        feature_tables_for_serving = self._load_feature_tables_for_serving(
            self.serving_environment, path
        )

        self.raw_model = _load_raw_model(path)

        # Reformat local metastore 3L tables to 2L. Non-local metastore 3L tables and 2L tables are unchanged.
        # This guarantees table name consistency between feature_spec and feature_tables_for_serving.
        # https://docs.google.com/document/d/1x_V9GshlnoAAFFCuDsXWdJVtop9MG2HWTUo5IK_1mEw
        original_feature_spec = FeatureSpec.load(path)
        # We call get_feature_spec_with_full_table_names to append the default metastore to 2L names,
        # as get_feature_spec_with_reformat_full_table_names expects full 3L table names and throws otherwise.
        # TODO (ML-26593): Consolidate this into a single function that allows either 2L/3L names.
        feature_spec_with_full_table_names = get_feature_spec_with_full_table_names(
            original_feature_spec
        )
        self.feature_spec = get_feature_spec_with_reformat_full_table_names(
            feature_spec_with_full_table_names
        )

        self.ft_metadata = self._get_ft_metadata(
            self.feature_spec, feature_tables_for_serving
        )
        self._validate_ft_metadata(self.ft_metadata)
        self.is_model_eligible_for_batch_lookup = (
            self._is_model_eligible_for_batch_lookup(self.ft_metadata)
        )

        if self.is_model_eligible_for_batch_lookup:
            self.batch_lookup_client = self._create_batch_lookup_client(
                self.ft_metadata
            )
        else:
            self.ft_to_lookup_client = self._create_lookup_clients(self.ft_metadata)

    # true if all feature tables using DynamoDB under same authorization (ie same region and keys)
    # this is true by default for customers using Sagemaker
    def _is_model_eligible_for_batch_lookup(
        self, ft_metadata: Dict[str, _FeatureTableMetadata]
    ):
        env = self._get_serving_environment()
        online_feature_tables = [meta.online_ft for _, meta in ft_metadata.items()]
        if env == mlflow_model_constants.SAGEMAKER:
            return is_primary_key_lookup(online_feature_tables)
        elif env == mlflow_model_constants.DATABRICKS:
            return tables_share_dynamodb_access_keys(
                online_feature_tables
            ) and is_primary_key_lookup(online_feature_tables)

        raise Exception(f"Internal Error: Unexpected serving_environment {env}.")

    @staticmethod
    def _load_feature_tables_for_serving(
        serving_environment: str, path: str
    ) -> AbstractFeatureTablesForServing:
        if serving_environment == mlflow_model_constants.DATABRICKS:
            return FeatureTablesForServing.load(
                path=os.getenv(FEATURE_TABLES_FOR_SERVING_FILEPATH_ENV)
            )
        elif serving_environment == mlflow_model_constants.SAGEMAKER:
            return FeatureTablesForSageMakerServing.load(path=path)

        raise Exception(
            f"Internal Error: Unexpected serving_environment {serving_environment}."
        )

    @staticmethod
    def _get_serving_environment() -> str:
        if os.environ.get(SERVING_ENVIRONMENT) == SAGEMAKER_SERVING_ENVIRONMENT:
            return mlflow_model_constants.SAGEMAKER
        return mlflow_model_constants.DATABRICKS

    @staticmethod
    def _is_lookup_client_instrumentation_enabled() -> bool:
        # environment variables stored as string, rather than boolean
        return os.getenv(LOOKUP_CLIENT_INSTRUMENTATION_ENABLED_ENV) == "true"

    @staticmethod
    def _check_support():
        if (
            databricks_utils.is_in_databricks_notebook()
            or databricks_utils.is_in_databricks_job()
        ):
            raise NotImplementedError(
                "Feature Store packaged models cannot be loaded with MLflow APIs. For batch "
                "inference, use FeatureStoreClient.score_batch."
            )

    @staticmethod
    def _validate_ft_metadata(ft_metadata):
        for (ft, meta) in ft_metadata.items():
            for lookup_key in meta.feature_col_infos_by_lookup_key.keys():
                if len(lookup_key) != len(meta.online_ft.primary_keys):
                    raise Exception(
                        f"Internal error: Online feature table has primary keys "
                        f"{meta.online_ft.primary_keys}, however FeatureSpec specifies "
                        f"{len(lookup_key)} lookup_keys: {lookup_key}."
                    )

    def _create_lookup_clients(self, ft_metadata):
        ft_to_lookup_client = {}
        for ft, meta in ft_metadata.items():
            ft_to_lookup_client[ft] = OnlineLookupClient(
                meta.online_ft, serving_environment=self.serving_environment
            )
        return ft_to_lookup_client

    def _create_batch_lookup_client(self, ft_metadata):
        online_fts = []
        for ft, meta in ft_metadata.items():
            online_fts.append(meta.online_ft)

        return OnlineLookupClient(
            online_fts, serving_environment=self.serving_environment
        )

    def predict(self, df: pd.DataFrame):
        self._validate_input(df)

        # if enabled, use metrics class to record metric values in child function calls
        output_metrics = None
        if self._is_lookup_client_instrumentation_enabled():
            output_metrics = LookupClientMetrics({})

        augment_with_features = lookup_call_maybe_with_metrics(
            self._augment_with_features, output_metrics, measuring_e2e_latency=True
        )

        model_input_df = augment_with_features(df, metrics=output_metrics)
        prediction = self.raw_model.predict(model_input_df)

        # will add any captured metrics to output via pandas series attributes
        if output_metrics:
            # if prediction is not pandas object (aka list or np array), convert prediction to panda series so it will have pandas attributes
            if not isinstance(prediction, pd.DataFrame) and not isinstance(
                prediction, pd.Series
            ):
                prediction = pd.Series(prediction)
            prediction.attrs = output_metrics.get_metrics()

        return prediction

    def _get_ft_metadata(
        self,
        feature_spec: FeatureSpec,
        fts_for_serving: AbstractFeatureTablesForServing,
    ) -> Dict[str, _FeatureTableMetadata]:
        ft_to_lookup_key_to_feature_col_infos = (
            self._group_fcis_by_feature_table_lookup_key(feature_spec)
        )
        ft_names = ft_to_lookup_key_to_feature_col_infos.keys()
        ft_to_online_ft = self._resolve_online_stores(
            feature_tables=ft_names, feature_tables_for_serving=fts_for_serving
        )

        return {
            ft: _FeatureTableMetadata(
                feature_col_infos_by_lookup_key=ft_to_lookup_key_to_feature_col_infos[
                    ft
                ],
                online_ft=ft_to_online_ft[ft],
            )
            for ft in ft_names
        }

    @staticmethod
    def _group_fcis_by_feature_table_lookup_key(
        feature_spec: FeatureSpec,
    ) -> Dict[str, LookupKeyToFeatureColumnInfosType]:
        """
        Re-organizes the provided FeatureSpec into a convenient structure for creating
        _FeatureTableMetadata objects.

        :return: Nested dictionary:
            {feature_table_name -> {lookup_key -> feature_column_infos}}
        """
        feature_table_to_lookup_key_to_fcis = defaultdict(lambda: defaultdict(list))
        for fci in feature_spec.feature_column_infos:
            feature_table_name = fci.table_name
            lookup_key = tuple(fci.lookup_key)
            feature_table_to_lookup_key_to_fcis[feature_table_name][lookup_key].append(
                fci
            )
        return feature_table_to_lookup_key_to_fcis

    def _get_overridden_feature_output_names(self, df: pd.DataFrame):
        """A feature value can be overridden in the provided DataFrame. There are two cases:
          1. The feature value is overridden for all rows in the df.
          2. The feature value is overridden for some but not all rows in the df.

        :return: Tuple<List[str], List[str]>
          (List of feature names with values that were fully overridden,
          list of feature names with values that were partially overridden)
        """
        overridden_feature_names = [
            fci.output_name
            for fci in self.feature_spec.feature_column_infos
            if fci.output_name in df.columns
        ]
        all_rows_overridden_idxs = df[overridden_feature_names].notna().all().values
        fully_overridden = [
            overridden_feature_names[i]
            for i in range(len(overridden_feature_names))
            if all_rows_overridden_idxs[i]
        ]
        partially_overridden = [
            name for name in overridden_feature_names if name not in fully_overridden
        ]
        return (fully_overridden, partially_overridden)

    def _validate_input(self, df: pd.DataFrame):
        """
        Validates:
            - df contains exactly one column per SourceDataColumnInfo
            - df contains exactly one column per lookup key in FeatureColumnInfos
            - df has no NaN lookup keys
        """
        req_source_column_names = [
            col_info.name for col_info in self.feature_spec.source_data_column_infos
        ]
        missing_source_columns = [
            col for col in req_source_column_names if col not in df.columns
        ]

        lookup_key_columns = set(
            itertools.chain.from_iterable(
                [
                    col_info.lookup_key
                    for col_info in self.feature_spec.feature_column_infos
                ]
            )
        )
        missing_lookup_key_columns = [
            col for col in lookup_key_columns if col not in df.columns
        ]

        missing_columns = missing_source_columns + missing_lookup_key_columns

        if missing_columns:
            raise ValueError(
                f"Input is missing columns '{missing_columns}'. "
                f"\n\tThe following lookup_key columns are required: {lookup_key_columns}."
                f"\n\tThe following columns are required for model input: {req_source_column_names}"
            )

        (
            fully_overridden,
            partially_overridden,
        ) = self._get_overridden_feature_output_names(df)

        df_column_name_counts = Counter(df.columns)
        no_dup_columns_allowed = (
            list(lookup_key_columns)
            + req_source_column_names
            + fully_overridden
            + partially_overridden
        )
        dup_columns = [
            col
            for col, count in df_column_name_counts.items()
            if count > 1 and col in no_dup_columns_allowed
        ]

        if dup_columns:
            raise ValueError(
                f"Input has duplicate columns: '{dup_columns}'"
                f"\n\tThe following column names must be unique: {no_dup_columns_allowed}"
            )

        lookup_key_df = df[list(lookup_key_columns)]
        cols_with_nulls = lookup_key_df.columns[lookup_key_df.isnull().any()].tolist()
        if cols_with_nulls:
            raise ValueError(
                f"Failed to lookup feature values due to null values for lookup_key columns "
                f"{cols_with_nulls}. The following columns cannot contain null values: "
                f"{lookup_key_columns}"
            )

    def _augment_with_features(
        self, df: pd.DataFrame, *, metrics: LookupClientMetrics = None
    ):
        """
        :param df: Pandas DataFrame provided by user as model input. This is expected to contain
        columns for each SourceColumnInfo, and for each lookup key of a FeatureColumnInfo. Columns
        with the same name as FeatureColumnInfo output_names will override those features, meaning
        they will not be queried from the online store.
        :return: Pandas DataFrame containing all features specified in the FeatureSpec, in order.
        """
        feature_dfs = []
        overridden_features_count = 0
        (
            fully_overridden_feature_output_names,
            partially_overridden_feature_output_names,
        ) = self._get_overridden_feature_output_names(df)

        pk_dfs = defaultdict(lambda: defaultdict(list))
        feature_column_infos_to_lookup_dict = defaultdict(lambda: defaultdict(list))
        lookup_clients = {}

        # Query online store(s) for feature values
        ft_meta: _FeatureTableMetadata
        for ft_name, ft_meta in self.ft_metadata.items():
            if not self.is_model_eligible_for_batch_lookup:
                lookup_clients[
                    ft_meta.online_ft.online_feature_table_name
                ] = self.ft_to_lookup_client[ft_name]

            # Iterate through the lookup_keys for this feature table, each of which is used to
            # lookup a list of features
            lookup_key: LookupKeyType
            feature_column_infos: List[FeatureColumnInfo]
            for (
                lookup_key,
                feature_column_infos,
            ) in ft_meta.feature_col_infos_by_lookup_key.items():
                # Do not lookup features that were fully overridden
                feature_column_infos_to_lookup = [
                    fci
                    for fci in feature_column_infos
                    if fci.output_name not in fully_overridden_feature_output_names
                ]
                if len(feature_column_infos_to_lookup) == 0:
                    # All features were overridden in the input DataFrame
                    continue

                pk_dfs[ft_meta.online_ft.online_feature_table_name][
                    lookup_key
                ] = self._get_primary_key_df(
                    lookup_key, ft_meta.online_ft.primary_keys, df
                )
                feature_column_infos_to_lookup_dict[
                    ft_meta.online_ft.online_feature_table_name
                ][lookup_key] = feature_column_infos_to_lookup

        feature_values_dfs = defaultdict(lambda: defaultdict(list))
        if self.is_model_eligible_for_batch_lookup and pk_dfs:
            feature_values_dfs = self._batch_lookup_and_rename_features(
                self.batch_lookup_client,
                pk_dfs,
                feature_column_infos_to_lookup_dict,
                metrics=metrics,
            )
        else:
            for oft_name, pk_df_by_lookup_key in pk_dfs.items():
                for lookup_key in pk_df_by_lookup_key.keys():
                    feature_values_dfs[oft_name][
                        lookup_key
                    ] = self._lookup_and_rename_features(
                        lookup_clients[oft_name],
                        pk_dfs[oft_name][lookup_key],
                        feature_column_infos_to_lookup_dict[oft_name][lookup_key],
                        metrics=metrics,
                    )

        for oft_name, feature_values_df_by_lookup_key in feature_values_dfs.items():
            for (
                lookup_key,
                feature_values_df,
            ) in feature_values_df_by_lookup_key.items():
                if feature_values_df.shape[0] != df.shape[0]:
                    raise Exception(
                        f"Internal Error: Expected {df.shape[0]} rows to be looked up from feature "
                        f"table {ft_name}, but found {feature_values_df.shape[0]}"
                    )

                # If any features were partially overridden, use the override values.
                # Filter all partially overridden feature output names down to those that
                # are in the feature table currently being processed.
                partially_overridden_feats = [
                    c
                    for c in feature_values_df.columns
                    if c in partially_overridden_feature_output_names
                ]
                if partially_overridden_feats:
                    # add partially overriden features to override count, count() returns the number of non-NaN values per column
                    # those are the values that will get overriden
                    overridden_features_count += (
                        df[partially_overridden_feats].count().sum()
                    )
                    # For each cell of overridden column, use the overridden value if provided, else
                    # the value looked up from the online store.
                    partially_overridden_feats_df = df[
                        partially_overridden_feats
                    ].combine_first(feature_values_df[partially_overridden_feats])
                    feature_values_df[
                        partially_overridden_feats
                    ] = partially_overridden_feats_df[partially_overridden_feats]
                feature_dfs.append(feature_values_df)

        # Include inputs from provided DataFrame
        source_data_df = df[
            [sdci.name for sdci in self.feature_spec.source_data_column_infos]
        ]
        # Renaming columns of source_data_df is not necessary because SourceDataColumnInfo's name
        # is always the same as the output_name.

        overridden_features_df = df[fully_overridden_feature_output_names]
        # add fully overriden features to override count
        overridden_features_count += overridden_features_df.count().sum()

        # Concatenate the following DataFrames, where N is the number of rows in `df`.
        #  1. feature_dfs - List of DataFrames, one per feature table looked up from the online
        #       store. Each DataFrame has N rows.
        #  2. source_data_df - DataFrame with N rows, containing SourceDataColumnInfo features
        #       from `df`.
        #  3. overridden_features_df - DataFrame with N rows, containing feature values from `df`
        #       that override FeatureColumnInfo features with non-null values for all rows.
        model_input_unordered = pd.concat(
            feature_dfs + [source_data_df, overridden_features_df], axis=1
        )

        if metrics:
            metrics.increase_metric(OVERRIDEN_FEATURE_COUNT, overridden_features_count)

        output_cols = [ci.output_name for ci in self.feature_spec.column_infos]
        model_input_df = model_input_unordered[output_cols]
        return model_input_df

    def _get_primary_key_df(
        self,
        lookup_key: LookupKeyType,
        primary_key: List[PrimaryKeyDetails],
        df: pd.DataFrame,
    ):
        """
        :return: A DataFrame containing a column for each column in `primary_keys`, and a row per
          row in `df`.
        """
        lookup_key_df = df[list(lookup_key)]

        # Update the lookup_key_df column names to be those of the feature table primary
        # key columns, rather than the names of columns from the source DataFrame. This is
        # required by the lookup_features interface.
        lookup_key_to_ft_pk = {
            lookup_name: feature_pk.name
            for (lookup_name, feature_pk) in zip(lookup_key, primary_key)
        }
        return lookup_key_df.rename(lookup_key_to_ft_pk, axis=1)

    def _lookup_and_rename_features(
        self,
        lookup_client: OnlineLookupClient,
        primary_key_df: pd.DataFrame,
        feature_column_infos: List[FeatureColumnInfo],
        *,
        metrics: LookupClientMetrics = None,
    ) -> pd.DataFrame:
        """
        Looks up features from a single feature table, then renames them. Feature metadata is
         specified via `feature_column_infos`.
        """
        feature_names = [fci.feature_name for fci in feature_column_infos]
        feature_values = lookup_client.lookup_features(
            primary_key_df, feature_names, metrics=metrics
        )
        feature_name_to_output_name = {
            fci.feature_name: fci.output_name for fci in feature_column_infos
        }
        return feature_values.rename(feature_name_to_output_name, axis=1)

    def _batch_lookup_and_rename_features(
        self,
        batch_lookup_client: OnlineLookupClient,
        primary_key_dfs: Dict[str, Dict[LookupKeyType, pd.DataFrame]],
        feature_column_infos_dict: Dict[
            str, Dict[LookupKeyType, List[FeatureColumnInfo]]
        ],
        *,
        metrics: LookupClientMetrics = None,
    ) -> Dict[str, Dict[LookupKeyType, pd.DataFrame]]:
        """
        Looks up features from all the feature tables in batch, then renames them. Feature metadata
         is specified via `feature_column_infos`.
        """
        feature_names = defaultdict(lambda: defaultdict(list))
        feature_name_to_output_names = defaultdict(lambda: defaultdict(list))
        for (
            oft_name,
            feature_column_infos_by_lookup_key,
        ) in feature_column_infos_dict.items():
            for (
                lookup_key,
                feature_column_infos,
            ) in feature_column_infos_by_lookup_key.items():
                feature_names[oft_name][lookup_key] = [
                    fci.feature_name for fci in feature_column_infos
                ]
                feature_name_to_output_names[oft_name][lookup_key] = {
                    fci.feature_name: fci.output_name for fci in feature_column_infos
                }
        feature_values = batch_lookup_client.batch_lookup_features(
            primary_key_dfs, feature_names, metrics=metrics
        )

        for oft_name, feature_values_by_lookup_key in feature_values.items():
            for lookup_key, _ in feature_values_by_lookup_key.items():
                feature_values[oft_name][lookup_key] = feature_values[oft_name][
                    lookup_key
                ].rename(feature_name_to_output_names[oft_name][lookup_key], axis=1)
        return feature_values

    def _resolve_online_stores(
        self,
        feature_tables: List[str],
        feature_tables_for_serving: AbstractFeatureTablesForServing,
    ) -> Dict[str, AbstractOnlineFeatureTable]:
        """
        :return: feature table name -> AbstractOnlineFeatureTable
        """
        all_fts_to_online_ft = {
            online_ft.feature_table_name: online_ft
            for online_ft in feature_tables_for_serving.online_feature_tables
        }

        missing = [ft for ft in feature_tables if ft not in all_fts_to_online_ft.keys()]
        if missing:
            raise Exception(
                f"Internal error: Online feature table information could not be found "
                f"for feature tables {missing}."
            )

        return {ft: all_fts_to_online_ft[ft] for ft in feature_tables}


def _load_pyfunc(path):
    """
    Called by ``pyfunc.load_pyfunc``.
    """
    if FEATURE_SERVING_RESPONSE_FORMAT_ENABLED:
        raw_model = _load_raw_model(path)
        try:
            py_model = raw_model._model_impl.python_model
            if (
                hasattr(py_model, "_is_databricks_internal_feature_serving_model")
                and py_model._is_databricks_internal_feature_serving_model
            ):
                from src.mlflowserving import scoring_server

                scoring_server.predictions_to_json = (
                    patch_feature_serving_predictions_to_json
                )
        except AttributeError as ae:
            # Shouldn't happen when the model is created with feature store client.
            print(f"Failed when checking model attribute: {repr(ae)}")
        except ImportError as ie:
            raise RuntimeError(f"Failed to load feature serving model: {repr(ie)}")

    return _FeatureStoreModelWrapper(path)



from typing import List, Optional

from databricks.feature_store.entities._feature_store_object import _FeatureStoreObject
from databricks.feature_store.entities.data_type import DataType
from databricks.feature_store.entities.online_store_for_serving import (
    AbstractOnlineStoreForServing,
    OnlineStoreForSageMakerServing,
    OnlineStoreForServing,
)
from databricks.feature_store.protos.feature_store_serving_pb2 import (
    FeatureDetails as ProtoFeatureDetails,
)
from databricks.feature_store.protos.feature_store_serving_pb2 import (
    OnlineFeatureTableForSageMakerServing as ProtoOnlineFeatureTableForSageMakerServing,
)
from databricks.feature_store.protos.feature_store_serving_pb2 import (
    PrimaryKeyDetails as ProtoPrimaryKeyDetails,
)
from databricks.feature_store.protos.feature_store_serving_pb2 import (
    TimestampKeyDetails as ProtoTimestampKeyDetails,
)


class PrimaryKeyDetails(_FeatureStoreObject):
    def __init__(self, name: str, data_type: DataType):
        self._name = name
        self._data_type = data_type

    @property
    def name(self):
        return self._name

    @property
    def data_type(self):
        return self._data_type

    def __str__(self):
        return "{" + self._name + "}"

    def __repr__(self):
        return str(self)


class TimestampKeyDetails(_FeatureStoreObject):
    def __init__(self, name: str, data_type: DataType):
        self._name = name
        self._data_type = data_type

    @property
    def name(self):
        return self._name

    @property
    def data_type(self):
        return self._data_type

    def __str__(self):
        return "{" + self._name + "}"

    def __repr__(self):
        return str(self)


class FeatureDetails(_FeatureStoreObject):
    def __init__(self, name: str, data_type: DataType, data_type_details: str = ""):
        self._name = name
        self._data_type = data_type
        self._data_type_details = data_type_details

    @property
    def name(self):
        return self._name

    @property
    def data_type(self):
        return self._data_type

    @property
    def data_type_details(self):
        return self._data_type_details


class AbstractOnlineFeatureTable(_FeatureStoreObject):
    def __init__(
        self,
        feature_table_name: str,
        online_feature_table_name: str,
        online_store: AbstractOnlineStoreForServing,
        primary_keys: List[PrimaryKeyDetails],
        feature_table_id: str,
        features: List[FeatureDetails],
        timestamp_keys: List[TimestampKeyDetails],
    ):
        self._feature_table_name = feature_table_name
        self._online_feature_table_name = online_feature_table_name
        self._online_store = online_store
        self._primary_keys = primary_keys
        self._feature_table_id = feature_table_id
        self._features = features
        self._timestamp_keys = timestamp_keys

    @property
    def feature_table_name(self):
        return self._feature_table_name

    @property
    def online_feature_table_name(self):
        return self._online_feature_table_name

    @property
    def online_store(self):
        return self._online_store

    @property
    def primary_keys(self):
        return self._primary_keys

    @property
    def feature_table_id(self):
        return self._feature_table_id

    @property
    def features(self):
        return self._features

    @property
    def timestamp_keys(self):
        return self._timestamp_keys


class OnlineFeatureTable(AbstractOnlineFeatureTable):
    def __init__(
        self,
        **kwargs,
    ):
        super().__init__(**kwargs)

    @classmethod
    def from_proto(cls, the_proto):
        return cls(
            feature_table_name=the_proto.feature_table_name,
            online_feature_table_name=the_proto.online_feature_table_name,
            online_store=OnlineStoreForServing.from_proto(the_proto.online_store),
            primary_keys=[
                PrimaryKeyDetails(primary_key.name, primary_key.data_type)
                for primary_key in the_proto.primary_keys
            ],
            feature_table_id=the_proto.feature_table_id,
            features=[
                FeatureDetails(
                    feature.name, feature.data_type, feature.data_type_details
                )
                for feature in the_proto.features
            ],
            timestamp_keys=[
                TimestampKeyDetails(timestamp_key.name, timestamp_key.data_type)
                for timestamp_key in the_proto.timestamp_keys
            ],
        )


class OnlineFeatureTableForSageMakerServing(AbstractOnlineFeatureTable):
    def __init__(
        self,
        **kwargs,
    ):
        super().__init__(**kwargs)

    def to_proto(self) -> ProtoOnlineFeatureTableForSageMakerServing:
        proto = ProtoOnlineFeatureTableForSageMakerServing(
            feature_table_name=self.feature_table_name,
            online_feature_table_name=self.online_feature_table_name,
            feature_table_id=self._feature_table_id,
        )
        proto.online_store.CopyFrom(self.online_store.to_proto())
        for pk in self.primary_keys:
            proto.primary_keys.append(
                ProtoPrimaryKeyDetails(name=pk.name, data_type=pk.data_type)
            )
        for feat in self.features:
            proto.features.append(
                ProtoFeatureDetails(
                    name=feat.name,
                    data_type=feat.data_type,
                    data_type_details=feat.data_type_details,
                )
            )
        for timestamp_key in self.timestamp_keys:
            proto.timestamp_keys.append(
                ProtoTimestampKeyDetails(
                    name=timestamp_key.name, data_type=timestamp_key.data_type
                )
            )
        return proto

    @classmethod
    def from_proto(cls, the_proto):
        return cls(
            feature_table_name=the_proto.feature_table_name,
            online_feature_table_name=the_proto.online_feature_table_name,
            online_store=OnlineStoreForSageMakerServing.from_proto(
                the_proto.online_store
            ),
            primary_keys=[
                PrimaryKeyDetails(primary_key.name, primary_key.data_type)
                for primary_key in the_proto.primary_keys
            ],
            feature_table_id=the_proto.feature_table_id,
            features=[
                FeatureDetails(
                    feature.name, feature.data_type, feature.data_type_details
                )
                for feature in the_proto.features
            ],
            timestamp_keys=[
                TimestampKeyDetails(timestamp_key.name, timestamp_key.data_type)
                for timestamp_key in the_proto.timestamp_keys
            ],
        )



def detect_coinstalled_clients():
    """
    Co-installation detection. The lookup client has nearly identical code in its version of
    databricks/feature_store/__init__.py.
    """

    # Detect if core client installed
    try:
        from databricks._feature_store_pkg_metadata import _core_client_pkg_metadata

        core_client_installed = True
    except ImportError:
        core_client_installed = False
    except Exception as e:
        print(
            f"Internal Warning: unexpected exception trying to import core client pkg_metadata: {e}"
        )
        core_client_installed = False

    # Detect if lookup client installed
    try:
        from databricks._feature_store_pkg_metadata import _lookup_client_pkg_metadata

        lookup_client_installed = True
    except ImportError:
        lookup_client_installed = False
    except Exception as e:
        print(
            f"Internal Warning: unexpected exception trying to import lookup client pkg_metadata: {e}"
        )
        lookup_client_installed = False

    # If neither client is installed, log a warning because we should never get into this situation.
    if not core_client_installed and not lookup_client_installed:
        print("Internal Warning: no feature store clients detected")

    # If both clients are installed, throw an exception because the earlier installed client will be in a broken
    # state due to having some of its files clobbered by the later installed client.  Note the order
    # of which clients are mentioned is different between clients to indicate which __init__.py emitted this error.
    if core_client_installed and lookup_client_installed:
        raise Exception(
            "The Databricks Lookup client and Databricks Feature Store client cannot be installed in the "
            "same python environment. Use pip to uninstall both clients, then pip install the client "
            "you intend to use."
        )


detect_coinstalled_clients()



""" Defines the LookupEngine class, which is used to perform lookups on online stores. This class
differs from Publish in that its actions are read-only.
"""

import abc
import collections
import functools
import logging
from typing import List

import numpy as np
import pandas as pd

from databricks.feature_store.entities.online_feature_table import OnlineFeatureTable
from databricks.feature_store.entities.online_store_for_serving import (
    MySqlConf,
    SqlServerConf,
)
from databricks.feature_store.utils.metrics_utils import LookupClientMetrics


class LookupEngine(abc.ABC):
    @abc.abstractmethod
    def lookup_features(
        self,
        lookup_df: pd.DataFrame,
        feature_names: List[str],
        *,
        metrics: LookupClientMetrics = None,
    ) -> pd.DataFrame:
        raise NotImplementedError

    @abc.abstractmethod
    def shutdown(self) -> None:
        raise NotImplementedError



""" Defines the type conversion classes from pandas to SQL and vice versa.
"""
from databricks.feature_store.entities.data_type import DataType
from databricks.feature_store.utils.converter_utils import ConverterFactory
from databricks.feature_store.utils.pandas_type_utils import (
    PandasBitBooleanConverter,
    PandasDecimalTypeConverter,
    PandasIdentityConverter,
    PandasStringArrayTypeConverter,
    PandasStringMapTypeConverter,
)

"""
All supported converters.
"""
BASIC_DATA_TYPE_CONVERTERS = {
    DataType.SHORT: PandasIdentityConverter,
    DataType.INTEGER: PandasIdentityConverter,
    DataType.LONG: PandasIdentityConverter,
    DataType.FLOAT: PandasIdentityConverter,
    DataType.DOUBLE: PandasIdentityConverter,
    DataType.BOOLEAN: PandasBitBooleanConverter,
    DataType.STRING: PandasIdentityConverter,
    DataType.TIMESTAMP: PandasIdentityConverter,
    DataType.DATE: PandasIdentityConverter,
    DataType.BINARY: PandasIdentityConverter,
}

COMPLEX_DATA_TYPE_CONVERTERS = {
    DataType.DECIMAL: PandasDecimalTypeConverter,
    DataType.ARRAY: PandasStringArrayTypeConverter,
    DataType.MAP: PandasStringMapTypeConverter,
}

SQL_DATA_TYPE_CONVERTER_FACTORY = ConverterFactory(
    BASIC_DATA_TYPE_CONVERTERS, COMPLEX_DATA_TYPE_CONVERTERS
)



""" Defines the LookupSqlEngine class, which is used to perform
lookups on SQL databases. This class differs from PublishSqlEngine in that
its actions are read-only, and uses a Python connector instead of Java.
"""

import abc
import functools
import logging
from contextlib import contextmanager
from typing import List

import numpy as np
import pandas as pd
import sqlalchemy

from databricks.feature_store.entities.online_feature_table import OnlineFeatureTable
from databricks.feature_store.lookup_engine.lookup_engine import LookupEngine
from databricks.feature_store.utils.metrics_utils import (
    NAN_FEATURE_COUNT,
    LookupClientMetrics,
    lookup_call_maybe_with_metrics,
)
from databricks.feature_store.utils.sql_type_utils import (
    SQL_DATA_TYPE_CONVERTER_FACTORY,
)

_logger = logging.getLogger(__name__)


class LookupSqlEngine(LookupEngine):
    INFORMATION_SCHEMA = "INFORMATION_SCHEMA"
    TABLES = "TABLES"
    TABLE_CATALOG = "TABLE_CATALOG"
    TABLE_SCHEMA = "TABLE_SCHEMA"
    TABLE_NAME = "TABLE_NAME"
    COLUMNS = "COLUMNS"
    COLUMN_NAME = "COLUMN_NAME"
    COLUMN_KEY = "COLUMN_KEY"
    DATA_TYPE = "DATA_TYPE"
    TABLE_CONSTRAINTS = "TABLE_CONSTRAINTS"
    CONSTRAINT_COLUMN_USAGE = "CONSTRAINT_COLUMN_USAGE"
    CONSTRAINT_TYPE = "CONSTRAINT_TYPE"
    CONSTRAINT_NAME = "CONSTRAINT_NAME"

    @abc.abstractmethod
    def __init__(
        self, online_feature_table: OnlineFeatureTable, ro_user: str, ro_password: str
    ):
        self.online_store = online_feature_table.online_store
        (
            self.database_name,
            self.table_name,
        ) = online_feature_table.online_feature_table_name.split(".")
        self.user = ro_user
        self.password = ro_password
        self.host = self.online_store.extra_configs.host
        self.port = self.online_store.extra_configs.port

        self.primary_keys = online_feature_table.primary_keys
        self.feature_names_to_feature_details = {
            f.name: f for f in online_feature_table.features
        }
        self.features_to_type_converter = {
            feature.name: SQL_DATA_TYPE_CONVERTER_FACTORY.get_converter(feature)
            for feature in online_feature_table.features
        }

        self._validate_online_feature_table()

    @contextmanager
    def _get_connection(self):
        # Everything between here and the yield statement will be executed in contextmanager.__enter__()

        # We create a SQL engine instance per required online table in an inference request and dispose it afterwards.
        # A single inference request runs queries sequentially, so a connection pool is not required or beneficial.
        # TODO (ML-23188): Investigate sharing a long lived SQL engine and connection pool across inference requests.
        engine = sqlalchemy.create_engine(
            self.engine_url, poolclass=sqlalchemy.pool.NullPool
        )
        connection = engine.connect()

        # When the caller invokes "with _get_connection() as x", the connection will be returned as "x"
        yield connection

        # Everything below here will be executed in contextmanager.__exit__()

        connection.close()
        # Disposing the engine prevents connections from hanging around in memory if a connection pool is used.
        engine.dispose()

    @property
    def engine_url(self) -> str:
        raise NotImplementedError

    def lookup_features(
        self,
        lookup_df: pd.DataFrame,
        feature_names: List[str],
        *,
        metrics: LookupClientMetrics = None,
    ) -> pd.DataFrame:
        pk_filter_phrase = " AND ".join(
            [f"{self._sql_safe_name(pk)} = :{pk}" for pk in lookup_df.columns]
        )
        select_feat_phrase = ", ".join(
            f"{self._sql_safe_name(f)}" for f in feature_names
        )
        query = sqlalchemy.sql.text(
            f"SELECT {select_feat_phrase} FROM {self._sql_safe_name(self.table_name)} WHERE {pk_filter_phrase}"
        )
        with self._get_connection() as sql_connection:
            sql_query = functools.partial(
                self._run_lookup_sql_query,
                sql_connection,
                query,
                feature_names,
                metrics,
            )
            # TODO: Optimize the SQL query by batching: https://databricks.atlassian.net/browse/ML-16636
            feature_df = lookup_df.apply(sql_query, axis=1, result_type="expand")
        feature_df.columns = feature_names
        for feature in feature_names:
            feature_df[feature] = feature_df[feature].map(
                lambda x: self.features_to_type_converter[feature].to_pandas(x)
                if x
                else x,
                na_action="ignore",
            )
        return feature_df

    def _validate_online_feature_table(
        self,
    ) -> None:
        # Validate the Python database connection specified by the database location in OnlineFeatureTable.
        # Throw operational error if new connection is invalid
        try:
            with self._get_connection() as sql_connection:
                # Validate the online feature table exists in online database as specified by the OnlineFeatureTable.
                if not self._database_contains_feature_table(sql_connection):
                    raise ValueError(
                        f"Table {self.table_name} does not exist in database {self.database_name}."
                    )

                    # Validate the online feature table has the same primary keys specified by the OnlineFeatureTable.
                if not self._database_contains_primary_keys(sql_connection):
                    raise ValueError(
                        f"Table {self.table_name} does not contain primary keys {self.primary_keys}."
                    )
        except Exception as e:
            raise ValueError(f"Connection could not be established: {str(e)}.")

    def shutdown(self) -> None:
        """
        Closes the database connection if it exists on the SQL lookup engine.
        :return:
        """
        self._close()

    def _close(self) -> None:
        """
        This is a no-op because a new sql connection is opened for each query and closed after the query executes.
        """
        pass

    def _run_lookup_sql_query(
        self, sql_connection, query, feature_names, metrics, lookup_row
    ):
        """
        This helper function executes a single SQL query .
        """
        query_params = lookup_row.to_dict()
        run_lookup = lookup_call_maybe_with_metrics(sql_connection.execute, metrics)
        sql_data = run_lookup(query, query_params)
        feat_values = sql_data.fetchall()

        if len(feat_values) == 0:
            _logger.warning(
                f"No feature values found in {self.table_name} for {query_params}."
            )
            nan_features = np.empty(len(feature_names))
            nan_features[:] = np.nan
            if metrics:
                # 0 nan feature count still needs to be recorded to propogate metrics for inference request
                metrics.increase_metric(NAN_FEATURE_COUNT, len(nan_features))

            return nan_features

        # Return the first result
        return feat_values[0]

    @classmethod
    def _sql_safe_name(cls, name):
        raise NotImplementedError

    def _database_contains_feature_table(self, sql_connection):
        raise NotImplementedError

    def _database_contains_primary_keys(self, sql_connection):
        raise NotImplementedError



import json

from mlflow.utils.proto_json_utils import (
    NumpyEncoder,
    _get_jsonable_obj,
    parse_tf_serving_input,
)


def patch_feature_serving_predictions_to_json(raw_predictions, output):
    predictions = _get_jsonable_obj(raw_predictions, pandas_orient="records")
    json.dump({"features": predictions}, output, cls=NumpyEncoder)



import base64
import calendar
import json
from abc import ABC, abstractmethod
from datetime import date
from decimal import Context, Decimal
from typing import Union

import numpy as np
import pandas as pd

from databricks.feature_store.utils.converter_utils import return_if_nan, return_if_none
from databricks.feature_store.utils.data_type_details_utils import (
    ELEMENT_TYPE,
    KEY_TYPE,
    VALUE_TYPE,
    get_data_type_from_details,
    parse_decimal_details,
)

"""
This file defines various type converters used across different online stores for online lookup. 
Each online store specifies a set of type converters through constructing a ConverterFactory instance.
Each type converter inherits either a stateless or stateful converter with the same two functions:

to_online_store(value)
# This function is used to convert a feature value in pandas df type to python native type 
# that can be understood by the underlying online store python SDK for primary key lookup.
# This function only needs to be implemented for data types that's supported as a valid primary key data type.

to_pandas(value)
# This function is used to convert a feature value in python native type to pandas df type. 
# This pandas df will then be used for model inference. The expected return type must match 
# the pyspark df -> pandas df conversion for the given data type to avoid training/scoring skew.
"""


class PandasStatelessConverter(ABC):
    @staticmethod
    @abstractmethod
    def to_online_store(value):
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def to_pandas(value):
        raise NotImplementedError


class PandasStatefulConverter(ABC):
    def __init__(self, details, get_converter_detailed):
        self._details = details
        self._get_converter_detailed = get_converter_detailed

    @abstractmethod
    def to_online_store(self, value):
        raise NotImplementedError

    @abstractmethod
    def to_pandas(self, value):
        raise NotImplementedError


class PandasIdentityConverter(PandasStatelessConverter):
    @staticmethod
    def to_online_store(value):
        return value

    @staticmethod
    @return_if_none
    @return_if_nan
    def to_pandas(value):
        return value


# Basic DataType Converters

# Boolean
class PandasBooleanTypeConverter(PandasIdentityConverter):
    pass


class PandasNumericBooleanConverter(PandasStatelessConverter):
    @staticmethod
    def to_online_store(value: bool) -> int:
        return int(value)

    @staticmethod
    @return_if_none
    @return_if_nan
    def to_pandas(value: Union[int, Decimal]) -> bool:
        if value != 0 and value != 1:
            raise ValueError("Unsupported value for bool: " + str(value))
        return bool(value)


class PandasBitBooleanConverter(PandasStatelessConverter):
    @staticmethod
    def to_online_store(value: bool) -> int:
        return int(value)

    @staticmethod
    @return_if_none
    @return_if_nan
    def to_pandas(value: str) -> bool:
        if ord(value) != 0 and ord(value) != 1:
            raise ValueError("Unsupported value for bool: " + str(value))
        return bool(ord(value))


# Binary
class PandasBinaryTypeConverter(PandasStatelessConverter):
    @staticmethod
    def to_online_store(value) -> bytearray:
        return bytearray(value)

    @staticmethod
    @return_if_none
    @return_if_nan
    # TODO value is boto3 object
    def to_pandas(value):
        return bytearray(value.value)


class PandasBase64BinaryTypeConverter(PandasStatelessConverter):
    @staticmethod
    def to_online_store(value: bytearray) -> str:
        return str(base64.b64encode(value), "utf-8")

    @staticmethod
    @return_if_none
    @return_if_nan
    def to_pandas(value: str) -> bytearray:
        return bytearray(base64.b64decode(value))


# Short
class PandasShortTypeConverter(PandasStatelessConverter):
    @staticmethod
    def to_online_store(value: np.int16) -> int:
        return int(value)

    @staticmethod
    @return_if_none
    @return_if_nan
    def to_pandas(value: Union[int, Decimal]) -> np.int16:
        # TODO (ML-20967): We currently return an np.int16 with best effort if there are no undefined values.
        #  However, if a np.nan is provided, we will return np.nan which is a np.float instead.
        return np.int16(value)


# Integer
class PandasIntTypeConverter(PandasStatelessConverter):
    @staticmethod
    def to_online_store(value: np.int32) -> int:
        return int(value)

    @staticmethod
    @return_if_none
    @return_if_nan
    def to_pandas(value: Union[int, Decimal]) -> np.int32:
        # TODO (ML-20967): We currently return an np.int32 with best effort if there are no undefined values.
        #  However, if a np.nan is provided, we will return np.nan which is a np.float instead.
        return np.int32(value)


# Long
class PandasLongTypeConverter(PandasStatelessConverter):
    @staticmethod
    def to_online_store(value: np.int64) -> int:
        return int(value)

    @staticmethod
    @return_if_none
    @return_if_nan
    def to_pandas(value: Union[int, Decimal]) -> np.int64:
        # TODO (ML-20967): We currently return an np.int64 with best effort if there are no undefined values.
        #  However, if a np.nan is provided, we will return np.nan which is a np.float instead.
        return np.int64(value)


# String
class PandasStringTypeConverter(PandasStatelessConverter):
    @staticmethod
    def to_online_store(value: str) -> str:
        return str(value)

    @staticmethod
    @return_if_none
    @return_if_nan
    def to_pandas(value: str) -> str:
        return value


# Float
class PandasFloatTypeConverter(PandasStatelessConverter):
    @staticmethod
    def to_online_store(value: np.float32) -> float:
        return float(value)

    @staticmethod
    @return_if_none
    @return_if_nan
    def to_pandas(value: Union[float, Decimal]) -> np.float32:
        return np.float32(value)


# Double
class PandasDoubleTypeConverter(PandasStatelessConverter):
    @staticmethod
    def to_online_store(value: np.float64) -> float:
        return float(value)

    @staticmethod
    @return_if_none
    @return_if_nan
    def to_pandas(value: Union[float, Decimal]) -> np.float64:
        return np.float64(value)


# Timestamp
# Us -> microseconds precision
class PandasEpochTimestampTypeUsConverter(PandasStatelessConverter):
    @staticmethod
    def to_online_store(value: str) -> int:
        # TODO (ML-24922): Investigate if the value type is str or pd.Timestamp
        dt = pd.Timestamp(value).floor("us")
        return int(dt.timestamp() * 1e6)

    @staticmethod
    @return_if_none
    @return_if_nan
    def to_pandas(value: int) -> pd.Timestamp:
        return pd.Timestamp(int(value), unit="us")


class PandasIsoTimestampTypeUsConverter(PandasStatelessConverter):
    @staticmethod
    def to_online_store(value: str) -> str:
        # Pandas (<1.4) Timestamp doesn't support timespec to force microsecond output, which is required for lookup.
        # e.g. pd.Timestamp("2022-06-20T15:30:45.000000").isoformat() outputs '2022-06-20T15:30:45'
        # TODO (ML-24922): Investigate if the value type is str or pd.Timestamp.
        #  Also, consider using native pd.Timestamp.isoformat with timespec when feasible.
        dt = pd.Timestamp(value).floor("us").to_pydatetime()
        return dt.isoformat(timespec="microseconds")

    @staticmethod
    @return_if_none
    @return_if_nan
    def to_pandas(value: str) -> pd.Timestamp:
        return pd.Timestamp(value).floor("us")


# Date
class PandasEpochDateTypeConverter(PandasStatelessConverter):
    """
    This converter will allow but truncate all inexact dates to the most recent date.
    """

    @staticmethod
    def to_online_store(value: str) -> int:
        dt = pd.Timestamp(value).date()
        return calendar.timegm(dt.timetuple())

    @staticmethod
    @return_if_none
    @return_if_nan
    def to_pandas(value: int) -> date:
        return pd.Timestamp(int(value), unit="s").date()


class PandasIsoDateTypeConverter(PandasStatelessConverter):
    @staticmethod
    def to_online_store(value: str) -> str:
        return pd.Timestamp(value).date().isoformat()

    @staticmethod
    @return_if_none
    @return_if_nan
    def to_pandas(value: str) -> date:
        return pd.Timestamp(value).date()


# Complex DataType Converters
# Decimal
class PandasDecimalTypeConverter(PandasStatefulConverter):
    def __init__(self, details, get_converter_detailed):
        super().__init__(details, get_converter_detailed)
        precision, _ = parse_decimal_details(details)
        self._precision = precision

    def to_online_store(self, value):
        raise NotImplementedError

    @return_if_none
    def to_pandas(self, value: Decimal) -> Decimal:
        # Set the Decimal context with the appropriate precision
        context = Context(prec=self._precision)
        return context.create_decimal(value)


# Array
class PandasArrayTypeConverter(PandasStatefulConverter):
    def __init__(self, details, get_converter_detailed):
        super().__init__(details, get_converter_detailed)
        element_data_type_details = details.get(ELEMENT_TYPE)
        element_data_type = get_data_type_from_details(element_data_type_details)
        element_converter = get_converter_detailed(
            element_data_type, details=element_data_type_details
        )
        self._element_converter = element_converter

    def to_online_store(self, value: np.ndarray):
        raise NotImplementedError

    @return_if_none
    def to_pandas(self, value: list) -> np.ndarray:
        if value is not None:
            return np.array(
                [
                    self._element_converter.to_pandas(x) if x is not None else x
                    for x in value
                ]
            )


class PandasStringArrayTypeConverter(PandasStatefulConverter):
    def __init__(self, details, get_converter_detailed):
        super().__init__(details, get_converter_detailed)

    def to_online_store(self, value: dict):
        raise NotImplementedError

    @return_if_none
    def to_pandas(self, value: str) -> dict:
        return json.loads(value)


# Map
class PandasMapTypeConverter(PandasStatefulConverter):
    def __init__(self, details, get_converter_detailed):
        super().__init__(details, get_converter_detailed)
        key_data_type_details = details.get(KEY_TYPE)
        key_data_type = get_data_type_from_details(key_data_type_details)
        value_data_type_details = details.get(VALUE_TYPE)
        value_data_type = get_data_type_from_details(value_data_type_details)
        key_converter = get_converter_detailed(
            key_data_type, details=key_data_type_details
        )
        value_converter = get_converter_detailed(
            value_data_type, details=value_data_type_details
        )
        self._key_converter = key_converter
        self._value_converter = value_converter

    def to_online_store(self, value: dict):
        raise NotImplementedError

    @return_if_none
    def to_pandas(self, value: dict) -> dict:
        if value is not None:
            return {
                self._key_converter.to_pandas(k): (
                    self._value_converter.to_pandas(v) if v is not None else v
                )
                for k, v in value.items()
            }


class PandasStringMapTypeConverter(PandasStatefulConverter):
    def __init__(self, details, get_converter_detailed):
        super().__init__(details, get_converter_detailed)

    def to_online_store(self, value: dict):
        raise NotImplementedError

    @return_if_none
    def to_pandas(self, value: str) -> dict:
        return json.loads(value)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: feature_spec.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from mlflow.protos.scalapb import scalapb_pb2 as scalapb_dot_scalapb__pb2
from mlflow.protos import databricks_pb2 as databricks__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12\x66\x65\x61ture_spec.proto\x12\x12\x66\x65\x61turestorecommon\x1a\x15scalapb/scalapb.proto\x1a\x10\x64\x61tabricks.proto\"$\n\x14SourceDataColumnInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\"\x84\x01\n\x11\x46\x65\x61tureColumnInfo\x12\x12\n\ntable_name\x18\x01 \x01(\t\x12\x14\n\x0c\x66\x65\x61ture_name\x18\x02 \x01(\t\x12\x12\n\nlookup_key\x18\x03 \x03(\t\x12\x13\n\x0boutput_name\x18\x04 \x01(\t\x12\x1c\n\x14timestamp_lookup_key\x18\x05 \x03(\t\"3\n\x0cInputBinding\x12\x11\n\tparameter\x18\x01 \x01(\t\x12\x10\n\x08\x62ound_to\x18\x02 \x01(\t\"u\n\x12OnDemandColumnInfo\x12\x10\n\x08udf_name\x18\x01 \x01(\t\x12\x38\n\x0einput_bindings\x18\x02 \x03(\x0b\x32 .featurestorecommon.InputBinding\x12\x13\n\x0boutput_name\x18\x03 \x01(\t\"\x87\x02\n\nColumnInfo\x12K\n\x17source_data_column_info\x18\x01 \x01(\x0b\x32(.featurestorecommon.SourceDataColumnInfoH\x00\x12\x44\n\x13\x66\x65\x61ture_column_info\x18\x02 \x01(\x0b\x32%.featurestorecommon.FeatureColumnInfoH\x00\x12G\n\x15on_demand_column_info\x18\x03 \x01(\x0b\x32&.featurestorecommon.OnDemandColumnInfoH\x00\x12\x15\n\x07include\x18\n \x01(\x08:\x04trueB\x06\n\x04info\"8\n\x10\x46\x65\x61tureTableInfo\x12\x12\n\ntable_name\x18\x01 \x01(\t\x12\x10\n\x08table_id\x18\x02 \x01(\t\"-\n\x0c\x46unctionInfo\x12\x10\n\x08udf_name\x18\x01 \x01(\t\x12\x0b\n\x03md5\x18\x02 \x01(\t\"\x96\x02\n\x0b\x46\x65\x61tureSpec\x12\x35\n\rinput_columns\x18\x01 \x03(\x0b\x32\x1e.featurestorecommon.ColumnInfo\x12\x1d\n\x15serialization_version\x18\x02 \x01(\x05\x12\x14\n\x0cworkspace_id\x18\x03 \x01(\x03\x12$\n\x1c\x66\x65\x61ture_store_client_version\x18\x04 \x01(\t\x12:\n\x0cinput_tables\x18\x05 \x03(\x0b\x32$.featurestorecommon.FeatureTableInfo\x12\x39\n\x0finput_functions\x18\x06 \x03(\x0b\x32 .featurestorecommon.FunctionInfoBC\n\'com.databricks.proto.featurestorecommonB\x10\x46\x65\x61tureSpecProto\xa0\x01\x01\xe2?\x02\x10\x01')



_SOURCEDATACOLUMNINFO = DESCRIPTOR.message_types_by_name['SourceDataColumnInfo']
_FEATURECOLUMNINFO = DESCRIPTOR.message_types_by_name['FeatureColumnInfo']
_INPUTBINDING = DESCRIPTOR.message_types_by_name['InputBinding']
_ONDEMANDCOLUMNINFO = DESCRIPTOR.message_types_by_name['OnDemandColumnInfo']
_COLUMNINFO = DESCRIPTOR.message_types_by_name['ColumnInfo']
_FEATURETABLEINFO = DESCRIPTOR.message_types_by_name['FeatureTableInfo']
_FUNCTIONINFO = DESCRIPTOR.message_types_by_name['FunctionInfo']
_FEATURESPEC = DESCRIPTOR.message_types_by_name['FeatureSpec']
SourceDataColumnInfo = _reflection.GeneratedProtocolMessageType('SourceDataColumnInfo', (_message.Message,), {
  'DESCRIPTOR' : _SOURCEDATACOLUMNINFO,
  '__module__' : 'feature_spec_pb2'
  # @@protoc_insertion_point(class_scope:featurestorecommon.SourceDataColumnInfo)
  })
_sym_db.RegisterMessage(SourceDataColumnInfo)

FeatureColumnInfo = _reflection.GeneratedProtocolMessageType('FeatureColumnInfo', (_message.Message,), {
  'DESCRIPTOR' : _FEATURECOLUMNINFO,
  '__module__' : 'feature_spec_pb2'
  # @@protoc_insertion_point(class_scope:featurestorecommon.FeatureColumnInfo)
  })
_sym_db.RegisterMessage(FeatureColumnInfo)

InputBinding = _reflection.GeneratedProtocolMessageType('InputBinding', (_message.Message,), {
  'DESCRIPTOR' : _INPUTBINDING,
  '__module__' : 'feature_spec_pb2'
  # @@protoc_insertion_point(class_scope:featurestorecommon.InputBinding)
  })
_sym_db.RegisterMessage(InputBinding)

OnDemandColumnInfo = _reflection.GeneratedProtocolMessageType('OnDemandColumnInfo', (_message.Message,), {
  'DESCRIPTOR' : _ONDEMANDCOLUMNINFO,
  '__module__' : 'feature_spec_pb2'
  # @@protoc_insertion_point(class_scope:featurestorecommon.OnDemandColumnInfo)
  })
_sym_db.RegisterMessage(OnDemandColumnInfo)

ColumnInfo = _reflection.GeneratedProtocolMessageType('ColumnInfo', (_message.Message,), {
  'DESCRIPTOR' : _COLUMNINFO,
  '__module__' : 'feature_spec_pb2'
  # @@protoc_insertion_point(class_scope:featurestorecommon.ColumnInfo)
  })
_sym_db.RegisterMessage(ColumnInfo)

FeatureTableInfo = _reflection.GeneratedProtocolMessageType('FeatureTableInfo', (_message.Message,), {
  'DESCRIPTOR' : _FEATURETABLEINFO,
  '__module__' : 'feature_spec_pb2'
  # @@protoc_insertion_point(class_scope:featurestorecommon.FeatureTableInfo)
  })
_sym_db.RegisterMessage(FeatureTableInfo)

FunctionInfo = _reflection.GeneratedProtocolMessageType('FunctionInfo', (_message.Message,), {
  'DESCRIPTOR' : _FUNCTIONINFO,
  '__module__' : 'feature_spec_pb2'
  # @@protoc_insertion_point(class_scope:featurestorecommon.FunctionInfo)
  })
_sym_db.RegisterMessage(FunctionInfo)

FeatureSpec = _reflection.GeneratedProtocolMessageType('FeatureSpec', (_message.Message,), {
  'DESCRIPTOR' : _FEATURESPEC,
  '__module__' : 'feature_spec_pb2'
  # @@protoc_insertion_point(class_scope:featurestorecommon.FeatureSpec)
  })
_sym_db.RegisterMessage(FeatureSpec)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\'com.databricks.proto.featurestorecommonB\020FeatureSpecProto\240\001\001\342?\002\020\001'
  _SOURCEDATACOLUMNINFO._serialized_start=83
  _SOURCEDATACOLUMNINFO._serialized_end=119
  _FEATURECOLUMNINFO._serialized_start=122
  _FEATURECOLUMNINFO._serialized_end=254
  _INPUTBINDING._serialized_start=256
  _INPUTBINDING._serialized_end=307
  _ONDEMANDCOLUMNINFO._serialized_start=309
  _ONDEMANDCOLUMNINFO._serialized_end=426
  _COLUMNINFO._serialized_start=429
  _COLUMNINFO._serialized_end=692
  _FEATURETABLEINFO._serialized_start=694
  _FEATURETABLEINFO._serialized_end=750
  _FUNCTIONINFO._serialized_start=752
  _FUNCTIONINFO._serialized_end=797
  _FEATURESPEC._serialized_start=800
  _FEATURESPEC._serialized_end=1078
# @@protoc_insertion_point(module_scope)



from typing import Union

from databricks.feature_store.entities._feature_store_object import _FeatureStoreObject
from databricks.feature_store.entities.feature_column_info import FeatureColumnInfo
from databricks.feature_store.entities.feature_spec_constants import (
    FEATURE_COLUMN_INFO,
    ON_DEMAND_COLUMN_INFO,
    SOURCE_DATA_COLUMN_INFO,
)
from databricks.feature_store.entities.on_demand_column_info import OnDemandColumnInfo
from databricks.feature_store.entities.source_data_column_info import (
    SourceDataColumnInfo,
)
from databricks.feature_store.protos.feature_spec_pb2 import (
    ColumnInfo as ProtoColumnInfo,
)


class ColumnInfo(_FeatureStoreObject):
    """
    ColumnInfo's structure and properties are mapped 1:1 to the ColumnInfo proto message, unless specified otherwise.
    """

    def __init__(
        self,
        info: Union[SourceDataColumnInfo, FeatureColumnInfo, OnDemandColumnInfo],
        include: bool,
    ):
        if not isinstance(
            info, (SourceDataColumnInfo, FeatureColumnInfo, OnDemandColumnInfo)
        ):
            raise ValueError(
                "info must be one of SourceDataColumnInfo, FeatureColumnInfo, OnDemandColumnInfo."
            )
        self._info = info
        self._include = include

    @property
    def info(
        self,
    ) -> Union[SourceDataColumnInfo, FeatureColumnInfo, OnDemandColumnInfo]:
        return self._info

    @property
    def include(self) -> bool:
        return self._include

    @property
    def output_name(self) -> str:
        """
        This field does not exist in the proto, and is provided for convenience.
        """
        return self.info.output_name

    @classmethod
    def from_proto(cls, column_info_proto):
        if column_info_proto.HasField(SOURCE_DATA_COLUMN_INFO):
            info = SourceDataColumnInfo.from_proto(
                column_info_proto.source_data_column_info
            )
        elif column_info_proto.HasField(FEATURE_COLUMN_INFO):
            info = FeatureColumnInfo.from_proto(column_info_proto.feature_column_info)
        elif column_info_proto.HasField(ON_DEMAND_COLUMN_INFO):
            info = OnDemandColumnInfo.from_proto(
                column_info_proto.on_demand_column_info
            )
        else:
            raise ValueError("Unsupported info type: " + str(column_info_proto))

        return ColumnInfo(info=info, include=column_info_proto.include)

    def to_proto(self):
        column_info = ProtoColumnInfo(include=self.include)
        if isinstance(self.info, SourceDataColumnInfo):
            column_info.source_data_column_info.CopyFrom(self.info.to_proto())
        elif isinstance(self.info, FeatureColumnInfo):
            column_info.feature_column_info.CopyFrom(self.info.to_proto())
        elif isinstance(self.info, OnDemandColumnInfo):
            column_info.on_demand_column_info.CopyFrom(self.info.to_proto())
        else:
            raise ValueError("Unsupported info type: " + str(self.info))

        return column_info



from typing import Any

from databricks.feature_store.entities._proto_enum_entity import _ProtoEnumEntity
from databricks.feature_store.protos.feature_store_serving_pb2 import (
    Cloud as ProtoCloud,
)


class Cloud(_ProtoEnumEntity):
    """Cloud types."""

    AWS = ProtoCloud.Value("AWS")
    AZURE = ProtoCloud.Value("AZURE")
    GCP = ProtoCloud.Value("GCP")

    @classmethod
    def _enum_type(cls) -> Any:
        return ProtoCloud
